# Copyright 2017 ARM Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# revision identifiers, used by Alembic.
revision = 'd2affd5b4172'
down_revision = 'f046346d1d87'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('volume_mapping',
                  sa.Column('auto_remove', sa.Boolean, nullable=True))



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""SQLAlchemy storage backend."""

from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import session as db_session
from oslo_db.sqlalchemy import utils as db_utils
from oslo_utils import importutils
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.orm import contains_eager
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import desc
from sqlalchemy.sql import func

from zun.common import consts
from zun.common import crypt
from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun.db.sqlalchemy import models

profiler_sqlalchemy = importutils.try_import('osprofiler.sqlalchemy')

CONF = zun.conf.CONF

_FACADE = None


def _create_facade_lazily():
    global _FACADE
    if _FACADE is None:
        # FIXME(hongbin): we need to remove reliance on autocommit semantics
        # ASAP since it's not compatible with SQLAlchemy 2.0
        db_session.enginefacade.configure(__autocommit=True)
        _FACADE = db_session.enginefacade.get_legacy_facade()
        if profiler_sqlalchemy:
            if CONF.profiler.enabled and CONF.profiler.trace_sqlalchemy:
                profiler_sqlalchemy.add_tracing(sa, _FACADE.get_engine(), "db")
    return _FACADE


def get_engine():
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(**kwargs):
    facade = _create_facade_lazily()
    return facade.get_session(**kwargs)


def get_backend():
    """The backend is this module itself."""
    return Connection()


def model_query(model, *args, **kwargs):
    """Query helper for simpler session usage.

    :param session: if present, the session to use
    """

    session = kwargs.get('session') or get_session()
    query = session.query(model, *args)
    return query


def add_identity_filter(query, value):
    """Adds an identity filter to a query.

    Filters results by ID, if supplied value is a valid integer.
    Otherwise attempts to filter results by UUID.

    :param query: Initial query to add filter to.
    :param value: Value for filtering results by.
    :return: Modified query.
    """
    if strutils.is_int_like(value):
        return query.filter_by(id=value)
    elif uuidutils.is_uuid_like(value):
        return query.filter_by(uuid=value)
    else:
        raise exception.InvalidIdentity(identity=value)


def _paginate_query(model, limit=None, marker=None, sort_key=None,
                    sort_dir=None, query=None, default_sort_key='id'):
    if not query:
        raise exception.ZunException('query cannot be none')
    sort_keys = [default_sort_key]
    if sort_key and sort_key not in sort_keys:
        sort_keys.insert(0, sort_key)
    try:
        query = db_utils.paginate_query(query, model, limit, sort_keys,
                                        marker=marker, sort_dir=sort_dir)
    except db_exc.InvalidSortKey:
        raise exception.InvalidParameterValue(
            _('The sort_key value "%(key)s" is an invalid field for sorting')
            % {'key': sort_key})
    return query.all()


class Connection(object):
    """SqlAlchemy connection."""

    def __init__(self):
        pass

    def _add_project_filters(self, context, query):
        if context.is_admin and context.all_projects:
            return query

        if context.project_id:
            query = query.filter_by(project_id=context.project_id)
        else:
            query = query.filter_by(user_id=context.user_id)

        return query

    def _add_filters(self, query, model, filters=None, filter_names=None):
        """Generic way to add filters to a Zun model"""
        if not filters:
            return query

        if not filter_names:
            filter_names = []

        for name in filter_names:
            if name in filters:
                value = filters[name]
                if isinstance(value, list):
                    column = getattr(model, name)
                    query = query.filter(column.in_(value))
                else:
                    column = getattr(model, name)
                    query = query.filter(column == value)

        return query

    def _add_containers_filters(self, query, filters):
        filter_names = ['name', 'image', 'project_id', 'user_id',
                        'memory', 'host', 'task_state', 'status',
                        'auto_remove', 'uuid', 'capsule_id']

        return self._add_filters(query, models.Container, filters=filters,
                                 filter_names=filter_names)

    def _add_container_type_filter(self, container_type, query):
        if container_type != consts.TYPE_ANY:
            query = query.filter_by(container_type=container_type)
        return query

    def list_containers(self, context, container_type, filters=None,
                        limit=None, marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.Container, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_container_type_filter(container_type, query)
            query = self._add_containers_filters(query, filters)
            return _paginate_query(models.Container, limit, marker,
                                   sort_key, sort_dir, query)

    def _validate_unique_container_name(self, context, name):
        session = get_session()
        with session.begin():
            if not CONF.compute.unique_container_name_scope:
                return
            lowername = name.lower()
            base_query = model_query(models.Container, session=session).\
                filter(func.lower(models.Container.name) == lowername)
            if CONF.compute.unique_container_name_scope == 'project':
                container_with_same_name = base_query.\
                    filter_by(project_id=context.project_id).count()
            elif CONF.compute.unique_container_name_scope == 'global':
                container_with_same_name = base_query.count()
            else:
                return

            if container_with_same_name > 0:
                raise exception.ContainerAlreadyExists(field='name',
                                                       value=lowername)

    def create_container(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new containers
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            if values.get('name'):
                self._validate_unique_container_name(context, values['name'])

            container = models.Container()
            container.update(values)
            try:
                container.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ContainerAlreadyExists(field='UUID',
                                                       value=values['uuid'])
            return container

    def get_container_by_uuid(self, context, container_type, container_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Container, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_container_type_filter(container_type, query)
            query = query.filter_by(uuid=container_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ContainerNotFound(container=container_uuid)

    def get_container_by_name(self, context, container_type, container_name):
        session = get_session()
        with session.begin():
            query = model_query(models.Container, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_container_type_filter(container_type, query)
            query = query.filter_by(name=container_name)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ContainerNotFound(container=container_name)
            except MultipleResultsFound:
                raise exception.Conflict('Multiple containers exist with same '
                                         'name. Please use the container uuid '
                                         'instead.')

    def destroy_container(self, context, container_type, container_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Container, session=session)
            query = self._add_container_type_filter(container_type, query)
            query = add_identity_filter(query, container_id)
            count = query.delete()
            if count != 1:
                raise exception.ContainerNotFound(container=container_id)

    def update_container(self, context, container_type, container_id, values):
        # NOTE(dtantsur): this can lead to very strange errors
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing Container.")
            raise exception.InvalidParameterValue(err=msg)

        if 'name' in values:
            self._validate_unique_container_name(context, values['name'])

        return self._do_update_container(container_type, container_id, values)

    def _do_update_container(self, container_type, container_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Container, session=session)
            query = self._add_container_type_filter(container_type, query)
            query = add_identity_filter(query, container_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ContainerNotFound(container=container_id)

            ref.update(values)
        return ref

    def _add_volume_mappings_filters(self, query, filters):
        filter_names = ['project_id', 'user_id', 'volume_id',
                        'container_path', 'container_uuid']
        return self._add_filters(query, models.VolumeMapping, filters=filters,
                                 filter_names=filter_names)

    def list_volume_mappings(self, context, filters=None, limit=None,
                             marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.VolumeMapping, session=session)
            query = query.join(models.Volume)
            query = self._add_project_filters(context, query)
            query = self._add_volume_filters(query, filters)
            query = self._add_volume_mappings_filters(query, filters)
            return _paginate_query(models.VolumeMapping, limit, marker,
                                   sort_key, sort_dir, query)

    def count_volume_mappings(self, context, **filters):
        session = get_session()
        with session.begin():
            query = model_query(models.VolumeMapping, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_volume_mappings_filters(query, filters)
            return query.count()

    def create_volume_mapping(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new volume_mappings
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            volume_mapping = models.VolumeMapping()
            volume_mapping.update(values)
            try:
                volume_mapping.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.VolumeMappingAlreadyExists(
                    field='UUID', value=values['uuid'])
            return volume_mapping

    def get_volume_mapping_by_uuid(self, context, volume_mapping_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.VolumeMapping, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(uuid=volume_mapping_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.VolumeMappingNotFound(
                    volume_mapping=volume_mapping_uuid)

    def destroy_volume_mapping(self, context, volume_mapping_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.VolumeMapping, session=session)
            query = add_identity_filter(query, volume_mapping_uuid)
            count = query.delete()
            if count != 1:
                raise exception.VolumeMappingNotFound(
                    volume_mapping_uuid)

    def update_volume_mapping(self, context, volume_mapping_uuid, values):
        # NOTE(dtantsur): this can lead to very strange errors
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing VolumeMapping.")
            raise exception.InvalidParameterValue(err=msg)

        return self._do_update_volume_mapping(volume_mapping_uuid, values)

    def _do_update_volume_mapping(self, volume_mapping_uuid, values):
        session = get_session()
        with session.begin():
            query = model_query(models.VolumeMapping, session=session)
            query = add_identity_filter(query, volume_mapping_uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.VolumeMappingNotFound(
                    volume_mapping=volume_mapping_uuid)

            ref.update(values)
        return ref

    def _add_volume_filters(self, query, filters):
        filter_names = ['project_id', 'user_id', 'cinder_volume_id',
                        'volume_provider']
        return self._add_filters(query, models.Volume, filters=filters,
                                 filter_names=filter_names)

    def create_volume(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new volume_mappings
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            volume = models.Volume()
            volume.update(values)
            try:
                volume.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.VolumeAlreadyExists(field='UUID',
                                                    value=values['uuid'])
            return volume

    def get_volume_by_id(self, context, volume_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Volume, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(id=volume_id)
            try:
                return query.one()
            except NoResultFound:
                raise exception.VolumeNotFound(volume=volume_id)

    def destroy_volume(self, context, volume_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Volume, session=session)
            query = add_identity_filter(query, volume_uuid)
            count = query.delete()
            if count != 1:
                raise exception.VolumeNotFound(volume=volume_uuid)

    def update_volume(self, context, volume_uuid, values):
        # NOTE(dtantsur): this can lead to very strange errors
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing Volume.")
            raise exception.InvalidParameterValue(err=msg)

        return self._do_update_volume(volume_uuid, values)

    def _do_update_volume(self, volume_uuid, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Volume, session=session)
            query = add_identity_filter(query, volume_uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.VolumeNotFound(volume=volume_uuid)

            ref.update(values)
        return ref

    def destroy_zun_service(self, host, binary):
        session = get_session()
        with session.begin():
            query = model_query(models.ZunService, session=session)
            query = query.filter_by(host=host, binary=binary)
            count = query.delete()
            if count != 1:
                raise exception.ZunServiceNotFound(host=host, binary=binary)

    def update_zun_service(self, host, binary, values):
        session = get_session()
        with session.begin():
            query = model_query(models.ZunService, session=session)
            query = query.filter_by(host=host, binary=binary)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ZunServiceNotFound(host=host, binary=binary)

            if 'report_count' in values:
                if values['report_count'] > ref.report_count:
                    ref.last_seen_up = timeutils.utcnow()

            ref.update(values)
        return ref

    def get_zun_service(self, host, binary):
        session = get_session()
        with session.begin():
            query = model_query(models.ZunService, session=session)
            query = query.filter_by(host=host, binary=binary)
            try:
                return query.one()
            except NoResultFound:
                return None

    def create_zun_service(self, values):
        session = get_session()
        with session.begin():
            zun_service = models.ZunService()
            zun_service.update(values)
            try:
                zun_service.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ZunServiceAlreadyExists(
                    host=zun_service.host, binary=zun_service.binary)
            return zun_service

    def _add_zun_service_filters(self, query, filters):
        filter_names = ['disabled', 'host', 'binary', 'project_id', 'user_id']
        return self._add_filters(query, models.ZunService, filters=filters,
                                 filter_names=filter_names)

    def list_zun_services(self, filters=None, limit=None, marker=None,
                          sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.ZunService, session=session)
            if filters:
                query = self._add_zun_service_filters(query, filters)

            return _paginate_query(models.ZunService, limit, marker,
                                   sort_key, sort_dir, query)

    def list_zun_services_by_binary(self, binary):
        session = get_session()
        with session.begin():
            query = model_query(models.ZunService, session=session)
            query = query.filter_by(binary=binary)
            return _paginate_query(models.ZunService, query=query)

    def destroy_image(self, context, uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Image, session=session)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ImageNotFound(image=uuid)

    def pull_image(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new images
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()
            image = models.Image()
            image.update(values)
            try:
                image.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ImageAlreadyExists(tag=values['tag'],
                                                   repo=values['repo'])
            return image

    def update_image(self, image_id, values):
        # NOTE(dtantsur): this can lead to very strange errors
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing Image.")
            raise exception.InvalidParameterValue(err=msg)
        return self._do_update_image(image_id, values)

    def _do_update_image(self, image_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Image, session=session)
            query = add_identity_filter(query, image_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ImageNotFound(image=image_id)

            ref.update(values)
        return ref

    def _add_image_filters(self, query, filters):
        filter_names = ['repo', 'project_id', 'user_id', 'size']
        return self._add_filters(query, models.Image, filters=filters,
                                 filter_names=filter_names)

    def list_images(self, context, filters=None, limit=None, marker=None,
                    sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.Image, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_image_filters(query, filters)
            return _paginate_query(models.Image, limit, marker, sort_key,
                                   sort_dir, query)

    def get_image_by_id(self, context, image_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Image, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(id=image_id)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ImageNotFound(image=image_id)

    def get_image_by_uuid(self, context, image_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Image, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(uuid=image_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ImageNotFound(image=image_uuid)

    def _add_resource_providers_filters(self, query, filters):
        filter_names = ['name', 'root_provider', 'parent_provider', 'can_host']
        return self._add_filters(query, models.ResourceProvider,
                                 filters=filters,
                                 filter_names=filter_names)

    def list_resource_providers(self, context, filters=None, limit=None,
                                marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceProvider, session=session)
            query = self._add_resource_providers_filters(query, filters)
            return _paginate_query(models.ResourceProvider, limit, marker,
                                   sort_key, sort_dir, query)

    def create_resource_provider(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new resource providers
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            resource_provider = models.ResourceProvider()
            resource_provider.update(values)
            try:
                resource_provider.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ResourceProviderAlreadyExists(
                    field='UUID', value=values['uuid'])
            return resource_provider

    def get_resource_provider(self, context, provider_ident):
        if uuidutils.is_uuid_like(provider_ident):
            return self._get_resource_provider_by_uuid(context, provider_ident)
        else:
            return self._get_resource_provider_by_name(context, provider_ident)

    def _get_resource_provider_by_uuid(self, context, provider_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceProvider, session=session)
            query = query.filter_by(uuid=provider_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ResourceProviderNotFound(
                    resource_provider=provider_uuid)

    def _get_resource_provider_by_name(self, context, provider_name):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceProvider, session=session)
            query = query.filter_by(name=provider_name)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ResourceProviderNotFound(
                    resource_provider=provider_name)
            except MultipleResultsFound:
                raise exception.Conflict('Multiple resource providers exist '
                                         'with same name. Please use the uuid '
                                         'instead.')

    def destroy_resource_provider(self, context, provider_id):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceProvider, session=session)
            query = add_identity_filter(query, provider_id)
            count = query.delete()
            if count != 1:
                raise exception.ResourceProviderNotFound(
                    resource_provider=provider_id)

    def update_resource_provider(self, context, provider_id, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing ResourceProvider.")
            raise exception.InvalidParameterValue(err=msg)

        return self._do_update_resource_provider(provider_id, values)

    def _do_update_resource_provider(self, provider_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceProvider, session=session)
            query = add_identity_filter(query, provider_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceProviderNotFound(
                    resource_provider=provider_id)

            ref.update(values)
        return ref

    def list_resource_classes(self, context, limit=None, marker=None,
                              sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceClass, session=session)
            return _paginate_query(models.ResourceClass, limit, marker,
                                   sort_key, sort_dir, query)

    def create_resource_class(self, context, values):
        session = get_session()
        with session.begin():
            resource = models.ResourceClass()
            resource.update(values)
            try:
                resource.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ResourceClassAlreadyExists(
                    field='uuid', value=values['uuid'])
            return resource

    def get_resource_class(self, context, resource_ident):
        if uuidutils.is_uuid_like(resource_ident):
            return self._get_resource_class_by_uuid(context, resource_ident)
        else:
            return self._get_resource_class_by_name(context, resource_ident)

    def _get_resource_class_by_uuid(self, context, resource_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceClass, session=session)
            query = query.filter_by(uuid=resource_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ResourceClassNotFound(
                    resource_class=resource_uuid)

    def _get_resource_class_by_name(self, context, resource_name):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceClass, session=session)
            query = query.filter_by(name=resource_name)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ResourceClassNotFound(
                    resource_class=resource_name)

    def destroy_resource_class(self, context, resource_id):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceClass, session=session)
            count = query.delete()
            if count != 1:
                raise exception.ResourceClassNotFound(
                    resource_class=str(resource_id))

    def update_resource_class(self, context, resource_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.ResourceClass, session=session)
            query = query.filter_by(id=resource_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceClassNotFound(
                    resource_class=resource_id)

            ref.update(values)
        return ref

    def _add_inventories_filters(self, query, filters):
        filter_names = ['resource_provider_id', 'resource_class_id', 'total',
                        'reserved', 'min_unit', 'max_unit', 'step_size',
                        'allocation_ratio', 'is_nested']
        return self._add_filters(query, models.Inventory, filters=filters,
                                 filter_names=filter_names)

    def list_inventories(self, context, filters=None, limit=None,
                         marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            session = get_session()
            query = model_query(models.Inventory, session=session)
            query = self._add_inventories_filters(query, filters)
            query = query.join(models.Inventory.resource_provider)
            query = query.options(
                contains_eager(models.Inventory.resource_provider))
            return _paginate_query(models.Inventory, limit, marker,
                                   sort_key, sort_dir, query)

    def create_inventory(self, context, provider_id, values):
        session = get_session()
        with session.begin():
            values['resource_provider_id'] = provider_id
            inventory = models.Inventory()
            inventory.update(values)
            try:
                inventory.save(session=session)
            except db_exc.DBDuplicateEntry as e:
                fields = {c: values[c] for c in e.columns}
                raise exception.UniqueConstraintViolated(fields=fields)
            return inventory

    def get_inventory(self, context, inventory_id):
        session = get_session()
        with session.begin():
            session = get_session()
            query = model_query(models.Inventory, session=session)
            query = query.join(models.Inventory.resource_provider)
            query = query.options(
                contains_eager(models.Inventory.resource_provider))
            query = query.filter_by(id=inventory_id)
            try:
                return query.one()
            except NoResultFound:
                raise exception.InventoryNotFound(inventory=inventory_id)

    def destroy_inventory(self, context, inventory_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Inventory, session=session)
            query = query.filter_by(id=inventory_id)
            count = query.delete()
            if count != 1:
                raise exception.InventoryNotFound(inventory=inventory_id)

    def update_inventory(self, context, inventory_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Inventory, session=session)
            query = query.filter_by(id=inventory_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.InventoryNotFound(inventory=inventory_id)

            ref.update(values)
        return ref

    def _add_allocations_filters(self, query, filters):
        filter_names = ['resource_provider_id', 'resource_class_id',
                        'consumer_id', 'used', 'is_nested']
        return self._add_filters(query, models.Allocation, filters=filters,
                                 filter_names=filter_names)

    def list_allocations(self, context, filters=None, limit=None,
                         marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.Allocation, session=session)
            query = self._add_allocations_filters(query, filters)
            query = query.join(models.Allocation.resource_provider)
            query = query.options(
                contains_eager(models.Allocation.resource_provider))
            return _paginate_query(models.Allocation, limit, marker,
                                   sort_key, sort_dir, query)

    def create_allocation(self, context, values):
        session = get_session()
        with session.begin():
            allocation = models.Allocation()
            allocation.update(values)
            try:
                allocation.save(session=session)
            except db_exc.DBDuplicateEntry as e:
                fields = {c: values[c] for c in e.columns}
                raise exception.UniqueConstraintViolated(fields=fields)
            return allocation

    def get_allocation(self, context, allocation_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Allocation, session=session)
            query = query.join(models.Allocation.resource_provider)
            query = query.options(
                contains_eager(models.Allocation.resource_provider))
            query = query.filter_by(id=allocation_id)
            try:
                return query.one()
            except NoResultFound:
                raise exception.AllocationNotFound(allocation=allocation_id)

    def destroy_allocation(self, context, allocation_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Allocation, session=session)
            query = query.filter_by(id=allocation_id)
            count = query.delete()
            if count != 1:
                raise exception.AllocationNotFound(allocation=allocation_id)

    def update_allocation(self, context, allocation_id, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Allocation, session=session)
            query = query.filter_by(id=allocation_id)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.AllocationNotFound(allocation=allocation_id)

            ref.update(values)
        return ref

    def _add_compute_nodes_filters(self, query, filters):
        filter_names = ['hostname', 'rp_uuid']
        return self._add_filters(query, models.ComputeNode, filters=filters,
                                 filter_names=filter_names)

    def list_compute_nodes(self, context, filters=None, limit=None,
                           marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.ComputeNode, session=session)
            query = self._add_compute_nodes_filters(query, filters)
            return _paginate_query(models.ComputeNode, limit, marker,
                                   sort_key, sort_dir, query,
                                   default_sort_key='uuid')

    def create_compute_node(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new compute nodes
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()
            if not values.get('rp_uuid'):
                values['rp_uuid'] = values['uuid']

            compute_node = models.ComputeNode()
            compute_node.update(values)
            try:
                compute_node.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ComputeNodeAlreadyExists(
                    field='UUID', value=values['uuid'])
            return compute_node

    def get_compute_node(self, context, node_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.ComputeNode, session=session)
            query = query.filter_by(uuid=node_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ComputeNodeNotFound(
                    compute_node=node_uuid)

    def get_compute_node_by_hostname(self, context, hostname):
        session = get_session()
        with session.begin():
            query = model_query(models.ComputeNode, session=session)
            query = query.filter_by(hostname=hostname)
            try:
                return query.one()
            except NoResultFound:
                raise exception.ComputeNodeNotFound(
                    compute_node=hostname)
            except MultipleResultsFound:
                raise exception.Conflict('Multiple compute nodes exist with '
                                         'same hostname. Please use the uuid '
                                         'instead.')

    def destroy_compute_node(self, context, node_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.ComputeNode, session=session)
            query = query.filter_by(uuid=node_uuid)
            count = query.delete()
            if count != 1:
                raise exception.ComputeNodeNotFound(
                    compute_node=node_uuid)

    def update_compute_node(self, context, node_uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing ComputeNode.")
            raise exception.InvalidParameterValue(err=msg)

        return self._do_update_compute_node(node_uuid, values)

    def _do_update_compute_node(self, node_uuid, values):
        session = get_session()
        with session.begin():
            query = model_query(models.ComputeNode, session=session)
            query = query.filter_by(uuid=node_uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ComputeNodeNotFound(
                    compute_node=node_uuid)

            ref.update(values)
        return ref

    def get_pci_device_by_addr(self, node_id, dev_addr):
        session = get_session()
        with session.begin():
            pci_dev_ref = model_query(models.PciDevice, session=session).\
                filter_by(compute_node_uuid=node_id).\
                filter_by(address=dev_addr).\
                first()
            if not pci_dev_ref:
                raise exception.PciDeviceNotFound(node_id=node_id,
                                                  address=dev_addr)
            return pci_dev_ref

    def get_pci_device_by_id(self, id):
        session = get_session()
        with session.begin():
            pci_dev_ref = model_query(models.PciDevice, session=session).\
                filter_by(id=id).\
                first()
            if not pci_dev_ref:
                raise exception.PciDeviceNotFoundById(id=id)
            return pci_dev_ref

    def get_all_pci_device_by_node(self, node_id):
        session = get_session()
        with session.begin():
            return model_query(models.PciDevice, session=session).\
                filter_by(compute_node_uuid=node_id).\
                all()

    def get_all_pci_device_by_parent_addr(self, node_id, parent_addr):
        session = get_session()
        with session.begin():
            return model_query(models.PciDevice, session=session).\
                filter_by(compute_node_uuid=node_id).\
                filter_by(parent_addr=parent_addr).\
                all()

    def get_all_pci_device_by_container_uuid(self, container_uuid):
        session = get_session()
        with session.begin():
            return model_query(models.PciDevice, session=session).\
                filter_by(status=consts.ALLOCATED).\
                filter_by(container_uuid=container_uuid).\
                all()

    def destroy_pci_device(self, node_id, address):
        session = get_session()
        with session.begin():
            query = model_query(models.PciDevice, session=session).\
                filter_by(compute_node_uuid=node_id).\
                filter_by(address=address)
            count = query.delete()
            if count != 1:
                raise exception.PciDeviceNotFound(node_id=node_id,
                                                  address=address)

    def update_pci_device(self, node_id, address, values):
        session = get_session()
        with session.begin():
            query = model_query(models.PciDevice, session=session).\
                filter_by(compute_node_uuid=node_id).\
                filter_by(address=address)
            if query.update(values) == 0:
                device = models.PciDevice()
                device.update(values)
                device.save(session=session)
            return query.one()

    def action_start(self, context, values):
        session = get_session()
        with session.begin():
            action = models.ContainerAction()
            action.update(values)
            action.save(session=session)
            return action

    def action_finish(self, context, values):
        session = get_session()
        with session.begin():
            query = model_query(models.ContainerAction, session=session).\
                filter_by(container_uuid=values['container_uuid']).\
                filter_by(request_id=values['request_id']).\
                filter_by(action=values['action'])
            if query.update(values) != 1:
                raise exception.ContainerActionNotFound(
                    request_id=values['request_id'],
                    container_uuid=values['container_uuid'])
            return query.one()

    def actions_get(self, context, container_uuid):
        """Get all container actions for the provided uuid."""
        session = get_session()
        with session.begin():
            query = model_query(models.ContainerAction, session=session).\
                filter_by(container_uuid=container_uuid)
            actions = _paginate_query(models.ContainerAction, sort_dir='desc',
                                      sort_key='created_at', query=query)

            return actions

    def action_get_by_request_id(self, context, container_uuid, request_id):
        """Get the action by request_id and given container."""
        action = self._action_get_by_request_id(context, container_uuid,
                                                request_id)
        return action

    def _action_get_by_request_id(self, context, container_uuid, request_id):
        session = get_session()
        with session.begin():
            result = model_query(models.ContainerAction, session=session).\
                filter_by(container_uuid=container_uuid).\
                filter_by(request_id=request_id).\
                first()
            return result

    def _action_get_last_created_by_container_uuid(self, context,
                                                   container_uuid):
        session = get_session()
        with session.begin():
            result = model_query(models.ContainerAction, session=session).\
                filter_by(container_uuid=container_uuid).\
                order_by(desc("created_at"), desc("id")).\
                first()
            return result

    def action_event_start(self, context, values):
        """Start an event on a container action."""
        session = get_session()
        with session.begin():
            action = self._action_get_by_request_id(context,
                                                    values['container_uuid'],
                                                    values['request_id'])

            # When zun-compute restarts, the request_id was different with
            # request_id recorded in ContainerAction, so we can't get the
            # original recode according to request_id. Try to get the last
            # created action so that init_container can continue to finish
            # the recovery action.
            if not action and not context.project_id:
                action = self._action_get_last_created_by_container_uuid(
                    context, values['container_uuid'])

            if not action:
                raise exception.ContainerActionNotFound(
                    request_id=values['request_id'],
                    container_uuid=values['container_uuid'])

            values['action_id'] = action['id']

            event = models.ContainerActionEvent()
            event.update(values)
            event.save(session=session)

            return event

    def action_event_finish(self, context, values):
        """Finish an event on a container action."""
        session = get_session()
        with session.begin():
            action = self._action_get_by_request_id(context,
                                                    values['container_uuid'],
                                                    values['request_id'])

            # When zun-compute restarts, the request_id was different with
            # request_id recorded in ContainerAction, so we can't get the
            # original recode according to request_id. Try to get the last
            # created action so that init_container can continue to finish
            # the recovery action.
            if not action and not context.project_id:
                action = self._action_get_last_created_by_container_uuid(
                    context, values['container_uuid'])

            if not action:
                raise exception.ContainerActionNotFound(
                    request_id=values['request_id'],
                    container_uuid=values['container_uuid'])

            event = model_query(models.ContainerActionEvent, session=session).\
                filter_by(action_id=action['id']).\
                filter_by(event=values['event']).\
                first()

            if not event:
                raise exception.ContainerActionEventNotFound(
                    action_id=action['id'], event=values['event'])

            event.update(values)
            event.save(session=session)

            return event

    def action_events_get(self, context, action_id):
        session = get_session()
        with session.begin():
            query = model_query(models.ContainerActionEvent, session=session).\
                filter_by(action_id=action_id)
            events = _paginate_query(models.ContainerActionEvent,
                                     sort_dir='desc',
                                     sort_key='created_at', query=query)
            return events

    def quota_create(self, context, project_id, resource, limit):
        quota_ref = models.Quota()
        quota_ref.project_id = project_id
        quota_ref.resource = resource
        quota_ref.hard_limit = limit
        session = get_session()
        with session.begin():
            try:
                quota_ref.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.QuotaAlreadyExists(project_id=project_id,
                                                   resource=resource)
            return quota_ref

    def quota_get(self, context, project_id, resource):
        session = get_session()
        with session.begin():
            query = model_query(models.Quota, session=session).\
                filter_by(project_id=project_id).\
                filter_by(resource=resource)
            result = query.first()
            if not result:
                raise exception.ProjectQuotaNotFound(project_id=project_id)
        return result

    def quota_get_all_by_project(self, context, project_id):
        session = get_session()
        with session.begin():
            rows = model_query(models.Quota, session=session).\
                filter_by(project_id=project_id).\
                all()
            result = {'project_id': project_id}
            for row in rows:
                result[row.resource] = row.hard_limit

        return result

    def quota_update(self, context, project_id, resource, limit):
        session = get_session()
        with session.begin():
            query = model_query(models.Quota, session=session).\
                filter_by(project_id=project_id).\
                filter_by(resource=resource)

            result = query.update({'hard_limit': limit})
            if not result:
                raise exception.ProjectQuotaNotFound(project_id=project_id)

    def quota_destroy(self, context, project_id, resource):
        session = get_session()
        with session.begin():
            query = model_query(models.Quota, session=session).\
                filter_by(project_id=project_id).\
                filter_by(resource=resource)
            query.delete()

    def quota_destroy_all_by_project(self, context, project_id):
        session = get_session()
        with session.begin():
            model_query(models.Quota, session=session).\
                filter_by(project_id=project_id).\
                delete()

            model_query(models.QuotaUsage, session=session).\
                filter_by(project_id=project_id).\
                delete()

    def quota_class_create(self, context, class_name, resource, limit):
        quota_class_ref = models.QuotaClass()
        quota_class_ref.class_name = class_name
        quota_class_ref.resource = resource
        quota_class_ref.hard_limit = limit
        session = get_session()
        with session.begin():
            quota_class_ref.save(session=session)
        return quota_class_ref

    def quota_class_get(self, context, class_name, resource):
        session = get_session()
        with session.begin():
            result = model_query(models.QuotaClass, session=session).\
                filter_by(class_name=class_name).\
                filter_by(resource=resource).\
                first()

        if not result:
            raise exception.QuotaClassNotFound(class_name=class_name)
        return result

    def quota_class_get_default(self, context):
        session = get_session()
        with session.begin():
            rows = model_query(models.QuotaClass, session=session).\
                filter_by(class_name=consts.DEFAULT_QUOTA_CLASS_NAME).\
                all()

            result = {'class_name': consts.DEFAULT_QUOTA_CLASS_NAME}
            for row in rows:
                result[row.resource] = row.hard_limit

        return result

    def quota_class_get_all_by_name(self, context, class_name):
        session = get_session()
        with session.begin():
            rows = model_query(models.QuotaClass, session=session).\
                filter_by(class_name=class_name).\
                all()

            result = {'class_name': class_name}
            for row in rows:
                result[row.resource] = row.hard_limit

        return result

    def quota_class_update(self, context, class_name, resource, limit):
        session = get_session()
        with session.begin():
            result = model_query(models.QuotaClass, session=session).\
                filter_by(class_name=class_name).\
                filter_by(resource=resource).\
                update({'hard_limit': limit})

            if not result:
                raise exception.QuotaClassNotFound(class_name=class_name)

    def quota_usage_get_all_by_project(self, context, project_id):
        session = get_session()
        with session.begin():
            rows = model_query(models.QuotaUsage, session=session).\
                filter_by(project_id=project_id).\
                all()

            result = {'project_id': project_id}
            for row in rows:
                result[row.resource] = dict(in_use=row.in_use,
                                            reserved=row.reserved)

            return result

    def _add_networks_filters(self, query, filters):
        filter_names = ['name', 'neutron_net_id', 'project_id', 'user_id',
                        'host']
        return self._add_filters(query, models.Network, filters=filters,
                                 filter_names=filter_names)

    def list_networks(self, context, filters=None, limit=None,
                      marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.Network, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_networks_filters(query, filters)
            return _paginate_query(models.Network, limit, marker,
                                   sort_key, sort_dir, query)

    def create_network(self, context, values):
        # ensure defaults are present for new networks
        session = get_session()
        with session.begin():
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            network = models.Network()
            network.update(values)
            try:
                network.save(session=session)
            except db_exc.DBDuplicateEntry as e:
                if 'neutron_net_id' in e.columns:
                    raise exception.NetworkAlreadyExists(
                        field='neutron_net_id', value=values['neutron_net_id'])
                else:
                    raise exception.NetworkAlreadyExists(
                        field='UUID', value=values['uuid'])
            return network

    def update_network(self, context, network_uuid, values):
        # NOTE(dtantsur): this can lead to very strange errors
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing docker network.")
            raise exception.InvalidParameterValue(err=msg)
        return self._do_update_network(network_uuid, values)

    def _do_update_network(self, network_uuid, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Network, session=session)
            query = add_identity_filter(query, network_uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.NetworkNotFound(network=network_uuid)

            ref.update(values)
        return ref

    def get_network_by_uuid(self, context, network_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Network)
            query = self._add_project_filters(context, query)
            query = query.filter_by(uuid=network_uuid)
            try:
                return query.one()
            except NoResultFound:
                raise exception.NetworkNotFound(network=network_uuid)

    def destroy_network(self, context, network_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Network, session=session)
            query = add_identity_filter(query, network_uuid)
            count = query.delete()
            if count != 1:
                raise exception.NetworkNotFound(network=network_uuid)

    def list_exec_instances(self, context, filters=None, limit=None,
                            marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.ExecInstance)
            query = self._add_exec_instances_filters(query, filters)
            return _paginate_query(models.ExecInstance, limit, marker,
                                   sort_key, sort_dir, query)

    def _add_exec_instances_filters(self, query, filters):
        filter_names = ['container_id', 'exec_id', 'token']
        return self._add_filters(query, models.ExecInstance, filters=filters,
                                 filter_names=filter_names)

    def create_exec_instance(self, context, values):
        session = get_session()
        with session.begin():
            exec_inst = models.ExecInstance()
            exec_inst.update(values)
            try:
                exec_inst.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.ExecInstanceAlreadyExists(
                    exec_id=values['exec_id'])
            return exec_inst

    def count_usage(self, context, container_type, project_id, flag):
        session = get_session()
        with session.begin():
            if flag == 'containers':
                project_query = session.query(
                    func.count(models.Container.id)). \
                    filter_by(project_id=project_id). \
                    filter_by(container_type=container_type)
            elif flag in ['disk', 'cpu', 'memory']:
                project_query = session.query(
                    func.sum(getattr(models.Container, flag))). \
                    filter_by(project_id=project_id). \
                    filter_by(container_type=container_type)

            return project_query.first()

    def _add_registries_filters(self, query, filters):
        filter_names = ['name', 'domain', 'username', 'project_id', 'user_id']
        return self._add_filters(query, models.Registry, filters=filters,
                                 filter_names=filter_names)

    def list_registries(self, context, filters=None, limit=None,
                        marker=None, sort_key=None, sort_dir=None):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = self._add_project_filters(context, query)
            query = self._add_registries_filters(query, filters)
            result = _paginate_query(models.Registry, limit, marker,
                                     sort_key, sort_dir, query)

        for row in result:
            row['password'] = crypt.decrypt(row['password'])
        return result

    def create_registry(self, context, values):
        session = get_session()
        with session.begin():
            # ensure defaults are present for new registries
            if not values.get('uuid'):
                values['uuid'] = uuidutils.generate_uuid()

            original_password = values.get('password')
            if original_password:
                values['password'] = crypt.encrypt(values.get('password'))

            registry = models.Registry()
            registry.update(values)
            try:
                registry.save(session=session)
            except db_exc.DBDuplicateEntry:
                raise exception.RegistryAlreadyExists(
                    field='UUID', value=values['uuid'])

        if original_password:
            # the password is encrypted but we want to return the original
            # password
            registry['password'] = original_password

        return registry

    def update_registry(self, context, registry_uuid, values):
        session = get_session()
        with session.begin():
            # NOTE(dtantsur): this can lead to very strange errors
            if 'uuid' in values:
                msg = _("Cannot overwrite UUID for an existing registry.")
                raise exception.InvalidParameterValue(err=msg)

            original_password = values.get('password')
            if original_password:
                values['password'] = crypt.encrypt(values.get('password'))

            updated = self._do_update_registry(registry_uuid, values)

        if original_password:
            # the password is encrypted but we want to return the original
            # password
            updated['password'] = original_password

        return updated

    def _do_update_registry(self, registry_uuid, values):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = add_identity_filter(query, registry_uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.RegistryNotFound(registry=registry_uuid)

            ref.update(values)
        return ref

    def get_registry_by_id(self, context, registry_id):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(id=registry_id)
            try:
                result = query.one()
            except NoResultFound:
                raise exception.RegistryNotFound(registry=registry_id)

        result['password'] = crypt.decrypt(result['password'])
        return result

    def get_registry_by_uuid(self, context, registry_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(uuid=registry_uuid)
            try:
                result = query.one()
            except NoResultFound:
                raise exception.RegistryNotFound(registry=registry_uuid)

        result['password'] = crypt.decrypt(result['password'])
        return result

    def get_registry_by_name(self, context, registry_name):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = self._add_project_filters(context, query)
            query = query.filter_by(name=registry_name)
            try:
                result = query.one()
            except NoResultFound:
                raise exception.RegistryNotFound(registry=registry_name)
            except MultipleResultsFound:
                raise exception.Conflict('Multiple registries exist with same '
                                         'name. Please use the registry uuid '
                                         'instead.')

        result['password'] = crypt.decrypt(result['password'])
        return result

    def destroy_registry(self, context, registry_uuid):
        session = get_session()
        with session.begin():
            query = model_query(models.Registry, session=session)
            query = add_identity_filter(query, registry_uuid)
            try:
                count = query.delete()
                if count != 1:
                    raise exception.RegistryNotFound(registry=registry_uuid)
            except db_exc.DBReferenceError:
                raise exception.Conflict('Failed to delete registry '
                                         '%(registry)s since it is in use.',
                                         registry=registry_uuid)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add rp_uuid to compute_node

Revision ID: d502ce8fb705
Revises: b2bda272f4dd
Create Date: 2019-08-25 15:27:06.626340

"""

# revision identifiers, used by Alembic.
revision = 'd502ce8fb705'
down_revision = 'b2bda272f4dd'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


COMPUTE_NODE_TABLE = sa.Table(
    'compute_node', sa.MetaData(),
    sa.Column('uuid', sa.String(36), primary_key=True, nullable=False),
    sa.Column('rp_uuid', sa.String(36), nullable=True))


def upgrade():
    op.add_column('compute_node',
                  sa.Column('rp_uuid', sa.String(length=36), nullable=True))
    op.create_unique_constraint('uniq_compute_node0rp_uuid',
                                'compute_node', ['rp_uuid'])

    # perform data migration between tables
    with sa.orm.Session(bind=op.get_bind()) as session:
        for row in session.query(COMPUTE_NODE_TABLE):
            session.execute(
                COMPUTE_NODE_TABLE.update().values(
                    rp_uuid=row.uuid).where(
                        COMPUTE_NODE_TABLE.c.uuid == row.uuid)
            )
        # this commit is necessary to allow further operations
        session.commit()

    op.alter_column('compute_node', 'rp_uuid',
                    nullable=False,
                    existing_type=sa.String(length=36),
                    existing_nullable=True,
                    existing_server_default=False)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""create exec_instance table

Revision ID: 26896d5f9053
Revises: 012a730926e8
Create Date: 2018-06-03 17:24:33.192354

"""

# revision identifiers, used by Alembic.
revision = '26896d5f9053'
down_revision = '012a730926e8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'exec_instance',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
        sa.Column('container_id', sa.Integer(), nullable=False),
        sa.Column('exec_id', sa.String(255), nullable=False),
        sa.Column('token', sa.String(255), nullable=True),
        sa.Column('url', sa.String(255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['container_id'], ['container.id'],
                                ondelete='CASCADE'),
        sa.UniqueConstraint('container_id', 'exec_id',
                            name='uniq_exec_instance0container_id_exec_id'),
    )



# Copyright (C) 2014 Juniper Networks, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet

from contrail_vrouter_api.vrouter_api import ContrailVRouterApi

from nova.network import linux_net
from nova import utils

from oslo_log import log as logging
from oslo_service import loopingcall

from nova.i18n import _


LOG = logging.getLogger(__name__)


class OpenContrailVIFDriver(object):
    def __init__(self):
        self._vrouter_semaphore = eventlet.semaphore.Semaphore()
        self._vrouter_client = ContrailVRouterApi(
            doconnect=True, semaphore=self._vrouter_semaphore)
        timer = loopingcall.FixedIntervalLoopingCall(self._keep_alive)
        timer.start(interval=2)

    def _keep_alive(self):
        self._vrouter_client.periodic_connection_check()

    def plug(self, instance, vif):
        vif_type = vif['type']

        LOG.debug('Plug vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        if_local_name = 'veth%s' % vif['id'][:8]
        if_remote_name = 'ns%s' % vif['id'][:8]

        # Device already exists so return.
        if linux_net.device_exists(if_local_name):
            return
        undo_mgr = utils.UndoManager()

        try:
            utils.execute('ip', 'link', 'add', if_local_name, 'type', 'veth',
                          'peer', 'name', if_remote_name, run_as_root=True)
            undo_mgr.undo_with(lambda: utils.execute(
                'ip', 'link', 'delete', if_local_name, run_as_root=True))

            utils.execute('ip', 'link', 'set', if_remote_name, 'address',
                          vif['address'], run_as_root=True)

        except Exception:
            LOG.exception("Failed to configure network")
            msg = _('Failed to setup the network, rolling back')
            undo_mgr.rollback_and_reraise(msg=msg, instance=instance)

    def attach(self, instance, vif, container_id):
        vif_type = vif['type']

        LOG.debug('Attach vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        if_local_name = 'veth%s' % vif['id'][:8]
        if_remote_name = 'ns%s' % vif['id'][:8]

        undo_mgr = utils.UndoManager()
        undo_mgr.undo_with(lambda: utils.execute(
            'ip', 'link', 'delete', if_local_name, run_as_root=True))
        ipv4_address, ipv4_netmask, ipv4_gateway = self._retrieve_ip_address(
            vif, 4)
        ipv6_address, ipv6_netmask, ipv6_gateway = self._retrieve_ip_address(
            vif, 6)
        ipv4_address = ipv4_address or '0.0.0.0'
        params = {
            'ip_address': ipv4_address,
            'vn_id': vif['network']['id'],
            'display_name': instance['display_name'],
            'hostname': instance['hostname'],
            'host': instance['host'],
            'vm_project_id': instance['project_id'],
            'port_type': 'NovaVMPort',
            'ip6_address': ipv6_address,
        }

        try:
            utils.execute('ip', 'link', 'set', if_remote_name, 'netns',
                          container_id, run_as_root=True)

            result = self._vrouter_client.add_port(
                instance['uuid'], vif['id'],
                if_local_name, vif['address'], **params)
            if not result:
                # follow the exception path
                raise RuntimeError(_('add_port returned %s') % str(result))
            utils.execute('ip', 'link', 'set', if_local_name, 'up',
                          run_as_root=True)
        except Exception:
            LOG.exception("Failed to attach the network")
            msg = _('Failed to attach the network, rolling back')
            undo_mgr.rollback_and_reraise(msg=msg, instance=instance)

        try:
            utils.execute('ip', 'netns', 'exec', container_id, 'ip', 'link',
                          'set', if_remote_name, 'address', vif['address'],
                          run_as_root=True)
            if ipv6_address:
                ip = ipv6_address + "/" + ipv6_netmask
                gateway = ipv6_gateway
                utils.execute('ip', 'netns', 'exec', container_id, 'ifconfig',
                              if_remote_name, 'inet6', 'add', ip,
                              run_as_root=True)
                utils.execute('ip', 'netns', 'exec', container_id, 'ip', '-6',
                              'route', 'replace', 'default', 'via', gateway,
                              'dev', if_remote_name, run_as_root=True)
            if ipv4_address != '0.0.0.0':
                ip = ipv4_address + "/" + ipv4_netmask
                gateway = ipv4_gateway
                utils.execute('ip', 'netns', 'exec', container_id, 'ifconfig',
                              if_remote_name, ip, run_as_root=True)
                utils.execute('ip', 'netns', 'exec', container_id, 'ip',
                              'route', 'replace', 'default', 'via', gateway,
                              'dev', if_remote_name, run_as_root=True)
            utils.execute('ip', 'netns', 'exec', container_id, 'ip', 'link',
                          'set', if_remote_name, 'up', run_as_root=True)
        except Exception:
            LOG.exception("Failed to attach vif", instance=instance)

    def _retrieve_ip_address(self, vif, version):
        address = None
        netmask = None
        gateway = None
        if 'subnets' in vif['network']:
            subnets = vif['network']['subnets']
            for subnet in subnets:
                ips = subnet['ips'][0]
                if (ips['version'] == version):
                    if ips['address'] is not None:
                        address = ips['address']
                        netmask = subnet['cidr'].split('/')[1]
                        gateway = subnet['gateway']['address']

        return address, netmask, gateway

    def unplug(self, instance, vif):
        vif_type = vif['type']
        if_local_name = 'veth%s' % vif['id'][:8]

        LOG.debug('Unplug vif_type=%(vif_type)s instance=%(instance)s '
                  'vif=%(vif)s',
                  {'vif_type': vif_type, 'instance': instance,
                   'vif': vif})

        try:
            self._vrouter_client.delete_port(vif['id'])
            if linux_net.device_exists(if_local_name):
                utils.execute('ip', 'link', 'delete', if_local_name,
                              run_as_root=True)
        except Exception:
            LOG.exception("Delete port failed", instance=instance)



#    Copyright 2017 Linaro Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import sys

from zun.common import config
from zun.common import service as zun_service
import zun.conf
from zun.websocket import websocketproxy

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def main():
    zun_service.prepare_service(sys.argv)
    config.parse_args(sys.argv)
    LOG.info("start websocket proxy")

    host = CONF.websocket_proxy.wsproxy_host
    port = CONF.websocket_proxy.wsproxy_port
    websocketproxy.ZunWebSocketProxy(
        listen_host=host,
        listen_port=port,
        cert=CONF.websocket_proxy.cert,
        key=CONF.websocket_proxy.key,
        ssl_only=CONF.websocket_proxy.ssl_only,
        file_only=True,
        RequestHandlerClass=websocketproxy.ZunProxyRequestHandler
    ).start_server()



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add runtime column

Revision ID: 945569b3669f
Revises: a251f1f61217
Create Date: 2017-08-04 09:10:47.810568

"""

# revision identifiers, used by Alembic.
revision = '945569b3669f'
down_revision = 'a251f1f61217'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('container',
                  sa.Column('runtime', sa.String(32), nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from zun.api.controllers import base
from zun.api.controllers.v1.schemas import quotas as schema
from zun.api import validation
from zun.common import exception
from zun.common import policy
from zun.common import quota
from zun import objects

QUOTAS = quota.QUOTAS


class QuotaController(base.Controller):
    """Controller for Quotas"""

    _custom_actions = {
        'defaults': ['GET'],
    }

    def _get_quotas(self, context, project_id, usages=False):
        values = QUOTAS.get_project_quotas(context, project_id,
                                           usages=usages)

        if usages:
            return values
        else:
            return {k: v['limit'] for k, v in values.items()}

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request, schema.query_param_update)
    @validation.validated(schema.query_param_update)
    def put(self, project_id, **quotas_dict):
        context = pecan.request.context
        policy.enforce(context, 'quota:update',
                       target={'project_id': project_id},
                       action='quota:update')
        for key, value in quotas_dict.items():
            value = int(value)
            quota = objects.Quota(context, project_id=project_id, resource=key,
                                  hard_limit=value)
            try:
                quota.create(context)
            except exception.QuotaAlreadyExists:
                quota.update(context)
        return self._get_quotas(context, project_id)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get(self, project_id, **kwargs):
        context = pecan.request.context
        usages = kwargs.get('usages', False)
        policy.enforce(context, 'quota:get',
                       target={'project_id': project_id},
                       action='quota:get')
        return self._get_quotas(context, project_id, usages=usages)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def defaults(self, project_id):
        context = pecan.request.context
        policy.enforce(context, 'quota:get_default',
                       target={'project_id': project_id},
                       action='quota:get_default')
        values = QUOTAS.get_defaults(context)
        return values

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def delete(self, project_id):
        context = pecan.request.context
        policy.enforce(context, 'quota:delete',
                       target={'project_id': project_id},
                       action='quota:delete')
        QUOTAS.destroy_all_by_project(context, project_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Scheduler
"""

from unittest import mock

from zun import objects
from zun.tests import base
from zun.tests.unit.scheduler import fakes


class SchedulerTestCase(base.TestCase):
    """Test case for base scheduler driver class."""

    driver_cls = fakes.FakeScheduler

    def setUp(self):
        super(SchedulerTestCase, self).setUp()
        self.driver = self.driver_cls()

    @mock.patch('zun.objects.ZunService.list_by_binary')
    @mock.patch('zun.api.servicegroup.ServiceGroup.service_is_up')
    def test_hosts_up(self, mock_service_is_up, mock_list_by_binary):
        service1 = objects.ZunService()
        service2 = objects.ZunService()
        service1.host = 'host1'
        service1.disabled = False
        service2.host = 'host2'
        service2.disabled = False
        services = [service1, service2]

        mock_list_by_binary.return_value = services
        mock_service_is_up.side_effect = [False, True]

        result = self.driver.hosts_up(self.context)
        self.assertEqual(['host2'], result)

        mock_list_by_binary.assert_called_once_with(self.context,
                                                    'zun-compute')
        calls = [mock.call(service1), mock.call(service2)]
        self.assertEqual(calls, mock_service_is_up.call_args_list)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from zun.api.controllers import link
from zun.common.policies import container as policies

_basic_keys = (
    'uuid',
    'user_id',
    'project_id',
    'name',
    'image',
    'links',
    'command',
    'status',
    'status_reason',
    'task_state',
    'cpu',
    'memory',
    'environment',
    'workdir',
    'ports',
    'hostname',
    'labels',
    'addresses',
    'image_pull_policy',
    'host',
    'restart_policy',
    'status_detail',
    'interactive',
    'tty',
    'image_driver',
    'security_groups',
    'auto_remove',
    'runtime',
    'disk',
    'auto_heal',
    'privileged',
    'healthcheck',
    'cpu_policy',
    'registry_id',
    'entrypoint',
    'created_at',
    'updated_at',
)


def format_container(context, url, container):
    def transform(key, value):
        if key not in _basic_keys:
            return
        # strip the key if it is not allowed by policy
        policy_action = policies.CONTAINER % ('get_one:%s' % key)
        if not context.can(policy_action, fatal=False, might_not_exist=True):
            return
        if key == 'uuid':
            yield ('uuid', value)
            if url:
                yield ('links', [link.make_link(
                    'self', url, 'containers', value),
                    link.make_link(
                        'bookmark', url,
                        'containers', value,
                        bookmark=True)])
        elif key == 'registry_id':
            if value:
                # the value is an internal id so replace it with the
                # user-facing uuid
                value = container.registry.uuid
            yield ('registry_id', value)
        else:
            yield (key, value)

    return dict(itertools.chain.from_iterable(
        transform(k, v) for k, v in container.as_dict().items()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for container service
"""

from oslo_db.sqlalchemy import models
from oslo_serialization import jsonutils as json
from oslo_utils import timeutils
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import sql
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.types import TypeDecorator, TEXT
from urllib import parse as urlparse

from zun.common import consts
import zun.conf


def MediumText():
    return Text().with_variant(MEDIUMTEXT(), 'mysql')


def table_args():
    engine_name = urlparse.urlparse(zun.conf.CONF.database.connection).scheme
    if engine_name == 'mysql':
        return {'mysql_engine': zun.conf.CONF.database.mysql_engine,
                'mysql_charset': "utf8"}
    return None


class JsonEncodedType(TypeDecorator):
    """Abstract base type serialized as json-encoded string in db."""
    type = None
    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is None:
            # Save default value according to current type to keep the
            # interface the consistent.
            value = self.type()
        elif not isinstance(value, self.type):
            raise TypeError(_("%s supposes to store %s objects, but %s given")
                            % (self.__class__.__name__,
                               self.type.__name__,
                               type(value).__name__))
        serialized_value = json.dumps(value)
        return serialized_value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class JSONEncodedDict(JsonEncodedType):
    """Represents dict serialized as json-encoded string in db."""
    type = dict


class JSONEncodedList(JsonEncodedType):
    """Represents list serialized as json-encoded string in db."""
    type = list


class ZunBase(models.TimestampMixin,
              models.ModelBase):

    metadata = None

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d

    def save(self, session=None):
        import zun.db.sqlalchemy.api as db_api

        if session is None:
            session = db_api.get_session()

        super(ZunBase, self).save(session)


Base = declarative_base(cls=ZunBase)


class ZunService(Base):
    """Represents health status of various zun services"""
    __tablename__ = 'zun_service'
    __table_args__ = (
        schema.UniqueConstraint("host", "binary",
                                name="uniq_zun_service0host0binary"),
        table_args()
    )

    id = Column(Integer, primary_key=True)
    host = Column(String(255))
    binary = Column(String(255))
    disabled = Column(Boolean, default=False)
    disabled_reason = Column(String(255))
    last_seen_up = Column(DateTime, nullable=True)
    forced_down = Column(Boolean, default=False)
    report_count = Column(Integer, nullable=False, default=0)
    availability_zone = Column(String(255), nullable=True)


class Container(Base):
    """Represents a container."""

    __tablename__ = 'container'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_container0uuid'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    project_id = Column(String(255))
    user_id = Column(String(255))
    uuid = Column(String(36))
    container_id = Column(String(36))
    name = Column(String(255))
    image = Column(String(255))
    cpu = Column(Float)
    cpu_policy = Column(String(255), default='shared')
    cpuset = Column(JSONEncodedDict, nullable=True)
    command = Column(JSONEncodedList)
    memory = Column(String(255))
    status = Column(String(20))
    status_reason = Column(Text, nullable=True)
    task_state = Column(String(20))
    environment = Column(JSONEncodedDict)
    workdir = Column(String(255))
    ports = Column(JSONEncodedList)
    hostname = Column(String(63))
    labels = Column(JSONEncodedDict)
    addresses = Column(JSONEncodedDict)
    image_pull_policy = Column(Text, nullable=True)
    host = Column(String(255))
    restart_policy = Column(JSONEncodedDict)
    status_detail = Column(String(50))
    interactive = Column(Boolean, default=False)
    tty = Column(Boolean, default=False)
    image_driver = Column(String(255))
    websocket_url = Column(String(255))
    websocket_token = Column(String(255))
    security_groups = Column(JSONEncodedList)
    auto_remove = Column(Boolean, default=False)
    runtime = Column(String(32))
    disk = Column(Integer, default=0)
    auto_heal = Column(Boolean, default=False)
    capsule_id = Column(Integer,
                        ForeignKey('container.id', ondelete='CASCADE'))
    started_at = Column(DateTime)
    privileged = Column(Boolean, default=False)
    healthcheck = Column(JSONEncodedDict)
    exposed_ports = Column(JSONEncodedDict)
    registry_id = Column(Integer,
                         ForeignKey('registry.id'),
                         nullable=True)
    container_type = Column(Integer, index=True,
                            default=consts.TYPE_CONTAINER,
                            server_default=str(consts.TYPE_CONTAINER))
    annotations = Column(MediumText())
    cni_metadata = Column(MediumText())
    entrypoint = Column(JSONEncodedList)


class VolumeMapping(Base):
    """Represents a volume mapping."""

    __tablename__ = 'volume_mapping'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_volume0uuid'),
        table_args()
    )
    uuid = Column(String(36), nullable=False)
    id = Column(Integer, primary_key=True, nullable=False)
    project_id = Column(String(255), nullable=True)
    user_id = Column(String(255), nullable=True)
    volume_id = Column(Integer,
                       ForeignKey('volume.id', ondelete="CASCADE"),
                       nullable=False)
    container_path = Column(String(255), nullable=True)
    container_uuid = Column(String(36), ForeignKey('container.uuid'))
    container = orm.relationship(
        Container,
        backref=orm.backref('volume'),
        foreign_keys=container_uuid,
        primaryjoin='and_(VolumeMapping.container_uuid==Container.uuid)')


class Volume(Base):
    """Represents a volume."""

    __tablename__ = 'volume'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_volume0uuid'),
        table_args()
    )
    uuid = Column(String(36), nullable=False)
    id = Column(Integer, primary_key=True, nullable=False)
    project_id = Column(String(255), nullable=True)
    user_id = Column(String(255), nullable=True)
    cinder_volume_id = Column(String(36), nullable=True)
    volume_provider = Column(String(36), nullable=False)
    connection_info = Column(MediumText())
    host = Column(String(255), nullable=True)
    auto_remove = Column(Boolean, default=False)
    contents = Column(MediumText())


class ExecInstance(Base):
    """Represents an exec instance."""

    __tablename__ = 'exec_instance'
    __table_args__ = (
        schema.UniqueConstraint(
            'container_id', 'exec_id',
            name='uniq_exec_instance0container_id_exec_id'),
        table_args()
    )
    id = Column(Integer, primary_key=True, nullable=False)
    container_id = Column(Integer,
                          ForeignKey('container.id', ondelete="CASCADE"),
                          nullable=False)
    exec_id = Column(String(255), nullable=False)
    token = Column(String(255), nullable=True)
    url = Column(String(255), nullable=True)
    container = orm.relationship(
        Container,
        backref=orm.backref('exec_instances'),
        foreign_keys=container_id,
        primaryjoin='and_(ExecInstance.container_id==Container.id)')


class Image(Base):
    """Represents an image. """

    __tablename__ = 'image'
    __table_args__ = (
        schema.UniqueConstraint(
            'repo', 'tag', 'host', name='uniq_image0repotaghost'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    project_id = Column(String(255))
    user_id = Column(String(255))
    uuid = Column(String(36))
    image_id = Column(String(255))
    repo = Column(String(255))
    tag = Column(String(255))
    size = Column(String(255))
    host = Column(String(255))


class ResourceProvider(Base):
    """Represents an resource provider. """

    __tablename__ = 'resource_provider'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_resource_provider0uuid'),
        table_args()
    )
    id = Column(Integer, primary_key=True, nullable=False)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=False)
    root_provider = Column(String(36), nullable=False)
    parent_provider = Column(String(36), nullable=True)
    can_host = Column(Integer, default=0)


class ResourceClass(Base):
    """Represents an resource class. """

    __tablename__ = 'resource_class'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_resource_class0uuid'),
        table_args()
    )
    id = Column(Integer, primary_key=True, nullable=False)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=False)


class Inventory(Base):
    """Represents an inventory. """

    __tablename__ = 'inventory'
    __table_args__ = (
        Index('inventory_resource_provider_id_idx',
              'resource_provider_id'),
        Index('inventory_resource_class_id_idx',
              'resource_class_id'),
        Index('inventory_resource_provider_resource_class_idx',
              'resource_provider_id', 'resource_class_id'),
        schema.UniqueConstraint(
            'resource_provider_id', 'resource_class_id',
            name='uniq_inventory0resource_provider_resource_class'),
        table_args()
    )
    id = Column(Integer, primary_key=True, nullable=False)
    resource_provider_id = Column(Integer, nullable=False)
    resource_class_id = Column(Integer, nullable=False)
    total = Column(Integer, nullable=False)
    reserved = Column(Integer, nullable=False)
    min_unit = Column(Integer, nullable=False)
    max_unit = Column(Integer, nullable=False)
    step_size = Column(Integer, nullable=False)
    allocation_ratio = Column(Float, nullable=False)
    is_nested = Column(Integer, nullable=False)
    blob = Column(JSONEncodedList)
    resource_provider = orm.relationship(
        "ResourceProvider",
        primaryjoin=('and_(Inventory.resource_provider_id == '
                     'ResourceProvider.id)'),
        foreign_keys=resource_provider_id)


class Allocation(Base):
    """Represents an allocation. """

    __tablename__ = 'allocation'
    __table_args__ = (
        Index('allocation_resource_provider_class_used_idx',
              'resource_provider_id', 'resource_class_id', 'used'),
        Index('allocation_resource_class_id_idx', 'resource_class_id'),
        Index('allocation_consumer_id_idx', 'consumer_id'),
        table_args()
    )
    id = Column(Integer, primary_key=True, nullable=False)
    resource_provider_id = Column(Integer, nullable=False)
    resource_class_id = Column(Integer, nullable=False)
    consumer_id = Column(String(36), nullable=False)
    used = Column(Integer, nullable=False)
    is_nested = Column(Integer, nullable=False)
    blob = Column(JSONEncodedList)
    resource_provider = orm.relationship(
        "ResourceProvider",
        primaryjoin=('and_(Allocation.resource_provider_id == '
                     'ResourceProvider.id)'),
        foreign_keys=resource_provider_id)


class ComputeNode(Base):
    """Represents a compute node. """

    __tablename__ = 'compute_node'
    __table_args__ = (
        schema.UniqueConstraint('rp_uuid', name='uniq_compute_node0rp_uuid'),
        table_args()
    )
    uuid = Column(String(36), primary_key=True, nullable=False)
    rp_uuid = Column(String(36), nullable=False)
    hostname = Column(String(255), nullable=False)
    numa_topology = Column(JSONEncodedDict, nullable=True)
    mem_total = Column(Integer, nullable=False, default=0)
    mem_free = Column(Integer, nullable=False, default=0)
    mem_available = Column(Integer, nullable=False, default=0)
    mem_used = Column(Integer, nullable=False, default=0)
    total_containers = Column(Integer, nullable=False, default=0)
    running_containers = Column(Integer, nullable=False, default=0)
    paused_containers = Column(Integer, nullable=False, default=0)
    stopped_containers = Column(Integer, nullable=False, default=0)
    cpus = Column(Integer, nullable=False, default=0)
    cpu_used = Column(Float, nullable=False, default=0.0)
    architecture = Column(String(32), nullable=True)
    os_type = Column(String(32), nullable=True)
    os = Column(String(64), nullable=True)
    kernel_version = Column(String(128), nullable=True)
    labels = Column(JSONEncodedDict)
    # Json string PCI Stats
    # '[{"vendor_id":"8086", "product_id":"1234", "count":3 }, ...]'
    pci_stats = Column(Text)
    disk_total = Column(Integer, nullable=False, default=0)
    disk_used = Column(Integer, nullable=False, default=0)
    disk_quota_supported = Column(Boolean, nullable=False, default=sql.false(),
                                  server_default=sql.false())
    runtimes = Column(JSONEncodedList, nullable=True)
    enable_cpu_pinning = Column(Boolean, nullable=False, default=sql.false(),
                                server_default=sql.false())


class PciDevice(Base):
    """Represents a PCI host device that can be passed through to instances."""

    __tablename__ = 'pci_device'
    __table_args__ = (
        Index('ix_pci_device_compute_node_uuid',
              'compute_node_uuid'),
        Index('ix_pci_device_container_uuid',
              'container_uuid'),
        Index('ix_pci_device_compute_node_uuid_parent_addr',
              'compute_node_uuid', 'parent_addr'),
        schema.UniqueConstraint(
            "compute_node_uuid", "address",
            name="uniq_pci_device0compute_node_uuid0address")
    )
    id = Column(Integer, primary_key=True)
    uuid = Column(String(36))
    compute_node_uuid = Column(String(36), ForeignKey('compute_node.uuid'),
                               nullable=False)

    # physical address of device domain:bus:slot.func (0000:09:01.1)
    address = Column(String(12), nullable=False)

    vendor_id = Column(String(4), nullable=False)
    product_id = Column(String(4), nullable=False)
    dev_type = Column(String(8), nullable=False)
    dev_id = Column(String(255))

    # label is abstract device name, that is used to unify devices with the
    # same functionality with different addresses or host.
    label = Column(String(255), nullable=False)

    status = Column(String(36), nullable=False)
    # the request_id is used to identify a device that is allocated for a
    # particular request
    request_id = Column(String(36), nullable=True)

    extra_info = Column(Text)

    container_uuid = Column(String(36))

    numa_node = Column(Integer, nullable=True)

    parent_addr = Column(String(12), nullable=True)
    container = orm.relationship(Container, backref="pci_device",
                                 foreign_keys=container_uuid,
                                 primaryjoin='and_('
                                 'PciDevice.container_uuid == Container.uuid)')


class ContainerAction(Base):
    """Represents a container action.

    The intention is that there will only be one of these pre user request. A
    lookup by(container_uuid, request_id) should always return a single result.
    """

    __tablename__ = 'container_actions'
    __table_args__ = (
        Index('container_uuid_idx', 'container_uuid'),
        Index('request_id_idx', 'request_id')
    )

    id = Column(Integer, primary_key=True, nullable=False)
    action = Column(String(255))
    container_uuid = Column(String(36),
                            ForeignKey('container.uuid', ondelete='CASCADE'),
                            nullable=False)
    request_id = Column(String(255))
    user_id = Column(String(255))
    project_id = Column(String(255))
    start_time = Column(DateTime, default=timeutils.utcnow)
    finish_time = Column(DateTime)
    message = Column(String(255))
    container = orm.relationship(
        "Container", backref="container_actions",
        foreign_keys=container_uuid,
        primaryjoin='and_(ContainerAction.container_uuid == Container.uuid)'
    )


class ContainerActionEvent(Base):
    """Track events that occur during an ContainerAction."""
    __tablename__ = 'container_actions_events'
    __table_args__ = ()

    id = Column(Integer, primary_key=True, nullable=False)
    event = Column(String(255))
    action_id = Column(Integer,
                       ForeignKey('container_actions.id', ondelete='CASCADE'),
                       nullable=False)
    start_time = Column(DateTime, default=timeutils.utcnow)
    finish_time = Column(DateTime)
    result = Column(String(255))
    traceback = Column(Text)
    details = Column(Text)


class Quota(Base):
    """Represents a single quota override for a project

    If there is no row for a given project id and resource, then the
    default for the quota class is used. If there is no row for a
    given quota class and resource, then the default for the deployment
    is used. If the row is present but the hard limit is None then the
    resource is unlimited.
    """

    __tablename__ = 'quotas'
    __table_args__ = ()
    id = Column(Integer, primary_key=True, nullable=False)
    project_id = Column(String(255), index=True)
    resource = Column(String(255), nullable=False)
    hard_limit = Column(Integer)


class QuotaClass(Base):
    """Represents a single quota override for a quota class

    If there is no row for a given quota class and resource, then the
    default for the deployment is used. If the row is present but the
    hard limit is None, then the resource is unlimited.
    """

    __tablename__ = 'quota_classes'
    __table_args__ = (
        Index('quota_classes_class_name_idx', 'class_name'),
    )
    id = Column(Integer, primary_key=True, nullable=False)
    class_name = Column(String(255))
    resource = Column(String(255))
    hard_limit = Column(Integer)


class QuotaUsage(Base):
    """Respents the current usage for a given resource."""

    __tablename__ = 'quota_usages'
    id = Column(Integer, primary_key=True)

    project_id = Column(String(255), index=True)
    resource = Column(String(255), nullable=False)

    in_use = Column(Integer, nullable=False)
    reserved = Column(Integer, nullable=False)

    @property
    def total(self):
        return self.in_use + self.reserved

    until_refresh = Column(Integer, nullable=True)


class Network(Base):
    """Represents a network. """

    __tablename__ = 'network'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_network0uuid'),
        schema.UniqueConstraint('neutron_net_id', 'host',
                                name='uniq_network0neutron_net_id_host'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    neutron_net_id = Column(String(255))
    host = Column(String(255))
    network_id = Column(String(255))
    project_id = Column(String(255))
    user_id = Column(String(255))
    uuid = Column(String(36))


class Registry(Base):
    """Represents a registry. """

    __tablename__ = 'registry'
    __table_args__ = (
        schema.UniqueConstraint('uuid', name='uniq_registry0uuid'),
        table_args()
    )
    id = Column(Integer, primary_key=True)
    project_id = Column(String(255))
    user_id = Column(String(255))
    uuid = Column(String(36))
    name = Column(String(255))
    domain = Column(String(255), nullable=False)
    username = Column(String(255))
    password = Column(String(255))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import traceback

from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_versionedobjects import fields

from zun.db import api as dbapi
from zun.objects import base

LOG = logging.getLogger(__name__)


@base.ZunObjectRegistry.register
class ContainerAction(base.ZunPersistentObject, base.ZunObject):

    # Version 1.0: Initial version
    # Version 1.1: Add uuid column.
    # Version 1.2: Remove uuid column.
    VERSION = '1.2'

    fields = {
        'id': fields.IntegerField(),
        'action': fields.StringField(nullable=True),
        'container_uuid': fields.UUIDField(nullable=True),
        'request_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'project_id': fields.StringField(nullable=True),
        'start_time': fields.DateTimeField(tzinfo_aware=False, nullable=True),
        'finish_time': fields.DateTimeField(tzinfo_aware=False, nullable=True),
        'message': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, action, db_action):
        for field in action.fields:
            setattr(action, field, getattr(db_action, field, None))

        action.obj_reset_changes()
        return action

    @staticmethod
    def _from_db_object_list(context, cls, db_objects):
        """Converts a list of database entities to a list of formal objects."""
        return [ContainerAction._from_db_object(context, cls(context), obj)
                for obj in db_objects]

    @staticmethod
    def pack_action_start(context, container_uuid, action_name):
        values = {'request_id': context.request_id,
                  'container_uuid': container_uuid,
                  'user_id': context.user_id,
                  'project_id': context.project_id,
                  'action': action_name,
                  'start_time': context.timestamp}
        return values

    @staticmethod
    def pack_action_finish(context, container_uuid, action_name,
                           exc_val=None, exc_tb=None):
        values = {'request_id': context.request_id,
                  'container_uuid': container_uuid,
                  'action': action_name,
                  'finish_time': timeutils.utcnow()}
        if exc_tb is not None:
            values['message'] = 'Error'
        return values

    @base.remotable_classmethod
    def get_by_request_id(cls, context, container_uuid, request_id):
        db_action = dbapi.action_get_by_request_id(context, container_uuid,
                                                   request_id)
        if db_action:
            return cls._from_db_object(context, cls(context), db_action)

    @base.remotable_classmethod
    def action_start(cls, context, container_uuid, action_name,
                     want_result=True):
        values = cls.pack_action_start(context, container_uuid, action_name)
        db_action = dbapi.action_start(context, values)
        if want_result:
            return cls._from_db_object(context, cls(context), db_action)

    @base.remotable_classmethod
    def action_finish(cls, context, container_uuid, action_name, exc_val=None,
                      exc_tb=None, want_result=True):
        values = cls.pack_action_finish(context, container_uuid, action_name,
                                        exc_val=exc_val, exc_tb=exc_tb)
        db_action = dbapi.action_finish(context, values)
        if want_result:
            return cls._from_db_object(context, cls(context), db_action)

    @base.remotable_classmethod
    def get_by_container_uuid(cls, context, container_uuid):
        db_actions = dbapi.actions_get(context, container_uuid)
        return ContainerAction._from_db_object_list(context, cls, db_actions)


@base.ZunObjectRegistry.register
class ContainerActionEvent(base.ZunPersistentObject, base.ZunObject):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'id': fields.IntegerField(),
        'event': fields.StringField(nullable=True),
        'action_id': fields.IntegerField(nullable=True),
        'start_time': fields.DateTimeField(tzinfo_aware=False, nullable=True),
        'finish_time': fields.DateTimeField(tzinfo_aware=False, nullable=True),
        'result': fields.StringField(nullable=True),
        'traceback': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, event, db_event):
        for field in event.fields:
            setattr(event, field, db_event[field])

        event.obj_reset_changes()
        return event

    @staticmethod
    def _from_db_object_list(context, cls, db_objects):
        """Converts a list of database entities to a list of formal objects."""
        return [ContainerActionEvent._from_db_object(context, cls(context),
                                                     obj)
                for obj in db_objects]

    @staticmethod
    def pack_action_event_start(context, container_uuid, event_name):
        values = {'event': event_name,
                  'container_uuid': container_uuid,
                  'request_id': context.request_id,
                  'start_time': timeutils.utcnow()}
        return values

    @staticmethod
    def pack_action_event_finish(context, container_uuid, event_name,
                                 exc_val=None, exc_tb=None):
        values = {'event': event_name,
                  'container_uuid': container_uuid,
                  'request_id': context.request_id,
                  'finish_time': timeutils.utcnow()}
        if exc_tb is None:
            values['result'] = 'Success'
        else:
            values['result'] = 'Error'
            values['message'] = exc_val
            values['traceback'] = exc_tb
        return values

    @base.remotable_classmethod
    def event_start(cls, context, container_uuid, event_name,
                    want_result=True):
        values = cls.pack_action_event_start(context, container_uuid,
                                             event_name)
        db_event = dbapi.action_event_start(context, values)
        if want_result:
            return cls._from_db_object(context, cls(context), db_event)

    @base.remotable_classmethod
    def event_finish(cls, context, container_uuid, event_name, exc_val=None,
                     exc_tb=None, want_result=None):
        if exc_val:
            exc_val = str(exc_val)
        if exc_tb and not isinstance(exc_tb, str):
            exc_tb = ''.join(traceback.format_tb(exc_tb))
        values = cls.pack_action_event_finish(context, container_uuid,
                                              event_name, exc_val=exc_val,
                                              exc_tb=exc_tb)
        db_event = dbapi.action_event_finish(context, values)
        if want_result:
            return cls._from_db_object(context, cls(context), db_event)

    @base.remotable_classmethod
    def get_by_action(cls, context, action_id):
        db_events = dbapi.action_events_get(context, action_id)
        return ContainerActionEvent._from_db_object_list(context, cls,
                                                         db_events)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Possible actions on an container.

Action should probably match a user intention at the API level. Because they
can be user visible that should help to avoid confusion. For that reason they
tent to maintain the casing sent to the API.

Maintaining a list of actions here should protect against inconsistencies when
they are used.
"""

CREATE = 'create'
DELETE = 'delete'
REBOOT = 'reboot'
REBUILD = 'rebuild'
STOP = 'stop'
START = 'start'
PAUSE = 'pause'
UNPAUSE = 'unpause'
EXEC_CMD = 'exec_cmd'
KILL = 'kill'
UPDATE = 'update'
CONTAINER_ATTACH = 'container_attach'
RESIZE = 'resize'
ADD_SECURITY_GROUP = 'add_security_group'
REMOVE_SECURITY_GROUP = 'remove_security_group'
PUT_ARCHIVE = 'put_archive'
COMMIT = 'commit'
NETWORK_DETACH = 'network_detach'
NETWORK_ATTACH = 'network_attach'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict
from unittest import mock

from docker import errors
from oslo_utils import units
from oslo_utils import uuidutils

from zun.common import consts
from zun.common import exception
from zun import conf
from zun.container.docker.driver import DockerDriver
from zun.container.docker import utils as docker_utils
from zun.objects.container import Container
from zun.tests.unit.container import base
from zun.tests.unit.db import utils
from zun.tests.unit.objects import utils as obj_utils

LSCPU_ON = """# The following is the parsable format, which can be fed to other
# programs. Each different item in every column has an unique ID
# starting from zero.
# Socket,CPU,Online
0,0,Y
0,8,Y"""

CONF = conf.CONF

_numa_node = {
    'id': 0,
    'cpuset': [8],
    'mem_available': 32768,
    'mem_total': 32768,
    'pinned_cpus': []
}

_numa_topo_spec = [_numa_node]


class TestDockerDriver(base.DriverTestCase):

    @mock.patch('zun.container.docker.driver.DockerDriver.'
                '_get_host_storage_info')
    def setUp(self, mock_get):
        super(TestDockerDriver, self).setUp()
        self.driver = DockerDriver()
        dfc_patcher = mock.patch.object(docker_utils, 'docker_client')
        docker_client = dfc_patcher.start()
        self.dfc_context_manager = docker_client.return_value
        self.mock_docker = mock.MagicMock()
        self.mock_default_container = obj_utils.get_test_container(
            self.context)
        self.dfc_context_manager.__enter__.return_value = self.mock_docker
        self.addCleanup(dfc_patcher.stop)

    def test_inspect_image_path_is_none(self):
        self.mock_docker.inspect_image = mock.Mock()
        mock_image = mock.MagicMock()
        self.driver.inspect_image(mock_image)
        self.mock_docker.inspect_image.assert_called_once_with(mock_image)

    def test_get_image(self):
        self.mock_docker.get_image = mock.Mock()
        self.driver.get_image(name='image_name')
        self.mock_docker.get_image.assert_called_once_with('image_name')

    @mock.patch('zun.image.glance.driver.GlanceDriver.delete_image_tar')
    def test_delete_image(self, mock_delete_image):
        self.mock_docker.inspect_image = mock.Mock(
            return_value={'RepoTags': ['ubuntu:1']})
        img_id = '1234'
        self.driver.delete_image(self.context, img_id, 'glance')
        self.mock_docker.inspect_image.assert_called_once_with(img_id)
        self.assertTrue(mock_delete_image.called)

    def test_load_image(self):
        self.mock_docker.load_image = mock.Mock()
        mock_open_file = mock.mock_open()
        with mock.patch('zun.container.docker.driver.open', mock_open_file):
            self.driver.load_image('test')
            self.mock_docker.load_image.assert_called_once_with(
                mock_open_file.return_value)

    def test_images(self):
        self.mock_docker.images = mock.Mock()
        self.driver.images(repo='test')
        self.mock_docker.images.assert_called_once_with('test', False)

    @mock.patch('neutronclient.v2_0.client.Client.create_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.expose_ports')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.neutron.NeutronAPI.create_or_update_port')
    @mock.patch('zun.common.utils.get_security_group_ids')
    @mock.patch('zun.objects.container.Container.save')
    def test_create_image_path_is_none_with_overlay2(
            self, mock_save,
            mock_get_security_group_ids,
            mock_create_or_update_port,
            mock_connect,
            mock_expose_ports,
            mock_create_security_group):
        self.mock_docker.create_host_config = mock.Mock(
            return_value={'Id1': 'val1', 'key2': 'val2'})
        self.mock_docker.create_container = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.create_networking_config = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        image = {'path': '', 'image': '', 'repo': 'test', 'tag': 'test'}
        mock_container = self.mock_default_container
        mock_container.status = 'Creating'
        mock_container.healthcheck = {}
        networks = [{'network': 'fake-network'}]
        volumes = {}
        fake_port = {'mac_address': 'fake_mac'}
        mock_create_or_update_port.return_value = ([], fake_port)
        mock_create_security_group.return_value = {
            'security_group': {'id': 'fake-id'}}
        # DockerDriver with supported storage driver - overlay2
        self.driver._host.sp_disk_quota = True
        self.driver._host.storage_driver = 'overlay2'
        result_container = self.driver.create(self.context, mock_container,
                                              image, networks, volumes)
        host_config = {}
        host_config['mem_limit'] = '512M'
        host_config['cpu_shares'] = 1024
        host_config['restart_policy'] = {'Name': 'no', 'MaximumRetryCount': 0}
        host_config['runtime'] = 'runc'
        host_config['binds'] = {}
        host_config['network_mode'] = 'fake-network'
        host_config['storage_opt'] = {'size': '20G'}
        host_config['privileged'] = False
        self.mock_docker.create_host_config.assert_called_once_with(
            **host_config)

        kwargs = {
            'name': '%sea8e2a25-2901-438d-8157-de7ffd68d051' %
                    consts.NAME_PREFIX,
            'command': ['fake_command'],
            'entrypoint': ['fake_entrypoint'],
            'environment': {'key1': 'val1', 'key2': 'val2'},
            'working_dir': '/home/ubuntu',
            'labels': {'key1': 'val1', 'key2': 'val2'},
            'host_config': {'Id1': 'val1', 'key2': 'val2'},
            'stdin_open': True,
            'tty': True,
            'hostname': 'testhost',
            'volumes': [],
            'networking_config': {'Id': 'val1', 'key1': 'val2'},
            'mac_address': 'fake_mac',
            'ports': [('80', 'tcp')],
        }
        self.mock_docker.create_container.assert_called_once_with(
            image['repo'] + ":" + image['tag'], **kwargs)
        self.assertEqual('val1', result_container.container_id)
        self.assertEqual(result_container.status,
                         consts.CREATED)

    @mock.patch('neutronclient.v2_0.client.Client.create_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.expose_ports')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.neutron.NeutronAPI.create_or_update_port')
    @mock.patch('zun.common.utils.get_security_group_ids')
    @mock.patch('zun.objects.container.Container.save')
    def test_create_image_path_is_none_with_devicemapper(
            self, mock_save,
            mock_get_security_group_ids,
            mock_create_or_update_port,
            mock_connect,
            mock_expose_ports,
            mock_create_security_group):
        self.mock_docker.create_host_config = mock.Mock(
            return_value={'Id1': 'val1', 'key2': 'val2'})
        self.mock_docker.create_container = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.create_networking_config = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        image = {'path': '', 'image': '', 'repo': 'test', 'tag': 'test'}
        mock_container = self.mock_default_container
        mock_container.status = 'Creating'
        mock_container.healthcheck = {}
        networks = [{'network': 'fake-network'}]
        volumes = {}
        fake_port = {'mac_address': 'fake_mac'}
        mock_create_or_update_port.return_value = ([], fake_port)
        mock_create_security_group.return_value = {
            'security_group': {'id': 'fake-id'}}
        # DockerDriver with supported storage driver - overlay2
        self.driver._host.sp_disk_quota = True
        self.driver._host.storage_driver = 'devicemapper'
        self.driver._host.default_base_size = 10
        result_container = self.driver.create(self.context, mock_container,
                                              image, networks, volumes)
        host_config = {}
        host_config['mem_limit'] = '512M'
        host_config['cpu_shares'] = 1024
        host_config['restart_policy'] = {'Name': 'no', 'MaximumRetryCount': 0}
        host_config['runtime'] = 'runc'
        host_config['binds'] = {}
        host_config['network_mode'] = 'fake-network'
        host_config['storage_opt'] = {'size': '20G'}
        host_config['privileged'] = False
        self.mock_docker.create_host_config.assert_called_once_with(
            **host_config)

        kwargs = {
            'name': '%sea8e2a25-2901-438d-8157-de7ffd68d051' %
                    consts.NAME_PREFIX,
            'command': ['fake_command'],
            'entrypoint': ['fake_entrypoint'],
            'environment': {'key1': 'val1', 'key2': 'val2'},
            'working_dir': '/home/ubuntu',
            'labels': {'key1': 'val1', 'key2': 'val2'},
            'host_config': {'Id1': 'val1', 'key2': 'val2'},
            'stdin_open': True,
            'tty': True,
            'hostname': 'testhost',
            'volumes': [],
            'networking_config': {'Id': 'val1', 'key1': 'val2'},
            'mac_address': 'fake_mac',
            'ports': [('80', 'tcp')],
        }
        self.mock_docker.create_container.assert_called_once_with(
            image['repo'] + ":" + image['tag'], **kwargs)
        self.assertEqual('val1', result_container.container_id)
        self.assertEqual(result_container.status,
                         consts.CREATED)

    @mock.patch('neutronclient.v2_0.client.Client.create_security_group')
    @mock.patch('zun.network.neutron.NeutronAPI.expose_ports')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.neutron.NeutronAPI.create_or_update_port')
    @mock.patch('zun.common.utils.get_security_group_ids')
    @mock.patch('zun.objects.container.Container.save')
    def test_create_docker_api_version_1_24(
            self, mock_save,
            mock_get_security_group_ids,
            mock_create_or_update_port,
            mock_connect,
            mock_expose_ports,
            mock_create_security_group):
        CONF.set_override("docker_remote_api_version", "1.24", "docker")
        self.mock_docker.create_host_config = mock.Mock(
            return_value={'Id1': 'val1', 'key2': 'val2'})
        self.mock_docker.create_container = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.create_networking_config = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        image = {'path': '', 'image': '', 'repo': 'test', 'tag': 'test'}
        mock_container = self.mock_default_container
        mock_container.status = 'Creating'
        mock_container.healthcheck = {}
        mock_container.runtime = None
        networks = [{'network': 'fake-network'}]
        volumes = {}
        fake_port = {'mac_address': 'fake_mac'}
        mock_create_or_update_port.return_value = ([], fake_port)
        mock_create_security_group.return_value = {
            'security_group': {'id': 'fake-id'}}
        result_container = self.driver.create(self.context, mock_container,
                                              image, networks, volumes)
        host_config = {}
        host_config['mem_limit'] = '512M'
        host_config['cpu_shares'] = 1024
        host_config['restart_policy'] = {'Name': 'no', 'MaximumRetryCount': 0}
        host_config['runtime'] = None
        host_config['binds'] = {}
        host_config['network_mode'] = 'fake-network'
        host_config['storage_opt'] = {'size': '20G'}
        host_config['privileged'] = False
        self.mock_docker.create_host_config.assert_called_once_with(
            **host_config)

        kwargs = {
            'name': '%sea8e2a25-2901-438d-8157-de7ffd68d051' %
                    consts.NAME_PREFIX,
            'command': ['fake_command'],
            'entrypoint': ['fake_entrypoint'],
            'environment': {'key1': 'val1', 'key2': 'val2'},
            'working_dir': '/home/ubuntu',
            'labels': {'key1': 'val1', 'key2': 'val2'},
            'host_config': {'Id1': 'val1', 'key2': 'val2'},
            'stdin_open': True,
            'tty': True,
            'hostname': 'testhost',
            'volumes': [],
            'networking_config': {'Id': 'val1', 'key1': 'val2'},
            'mac_address': 'fake_mac',
            'ports': [('80', 'tcp')],
        }
        self.mock_docker.create_container.assert_called_once_with(
            image['repo'] + ":" + image['tag'], **kwargs)
        self.assertEqual('val1', result_container.container_id)
        self.assertEqual(result_container.status,
                         consts.CREATED)

    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.neutron.NeutronAPI.create_or_update_port')
    @mock.patch('zun.common.utils.get_security_group_ids')
    @mock.patch('zun.objects.container.Container.save')
    def test_create_docker_api_version_1_24_runtime_not_supported(
            self, mock_save,
            mock_get_security_group_ids,
            mock_create_or_update_port,
            mock_connect):
        CONF.set_override("docker_remote_api_version", "1.24", "docker")
        self.mock_docker.create_host_config = mock.Mock(
            return_value={'Id1': 'val1', 'key2': 'val2'})
        self.mock_docker.create_container = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.create_networking_config = mock.Mock(
            return_value={'Id': 'val1', 'key1': 'val2'})
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        image = {'path': '', 'image': '', 'repo': 'test', 'tag': 'test'}
        mock_container = self.mock_default_container
        mock_container.status = 'Creating'
        mock_container.runtime = 'runc'
        networks = [{'network': 'fake-network'}]
        volumes = {}
        fake_port = {'mac_address': 'fake_mac'}
        mock_create_or_update_port.return_value = ([], fake_port)
        with self.assertRaisesRegex(
                exception.ZunException,
                "Specifying runtime in Docker API is not supported"):
            self.driver.create(self.context, mock_container,
                               image, networks, volumes)

        self.mock_docker.create_host_config.assert_not_called()
        self.mock_docker.create_container.assert_not_called()

    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_network_for_container')
    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_exposed_ports')
    def test_delete_success(self, mock_cleanup_network_for_container,
                            mock_cleanup_exposed_ports):
        self.mock_docker.remove_container = mock.Mock()
        mock_container = self.mock_default_container
        self.driver.delete(self.context, mock_container, True)
        self.assertTrue(mock_cleanup_network_for_container.called)
        self.assertTrue(mock_cleanup_exposed_ports.called)
        self.mock_docker.remove_container.assert_called_once_with(
            mock_container.container_id, force=True)

    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_network_for_container')
    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_exposed_ports')
    def test_delete_fail_no_result(self, mock_cleanup_network_for_container,
                                   mock_cleanup_exposed_ports):
        with mock.patch.object(errors.APIError, '__str__',
                               return_value='404 Not Found') as mock_init:
            self.mock_docker.remove_container = mock.Mock(
                side_effect=errors.APIError('Error', '', ''))
            mock_container = mock.MagicMock()
            self.driver.delete(self.context, mock_container, True)
            self.assertTrue(mock_cleanup_network_for_container.called)
            self.assertTrue(mock_cleanup_exposed_ports.called)
            self.mock_docker.remove_container.assert_called_once_with(
                mock_container.container_id, force=True)
            self.assertEqual(1, mock_init.call_count)

    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_network_for_container')
    @mock.patch('zun.container.docker.driver.DockerDriver'
                '._cleanup_exposed_ports')
    def test_delete_fail_raise_error(self, mock_cleanup_network_for_container,
                                     mock_cleanup_exposed_ports):
        with mock.patch.object(errors.APIError, '__str__',
                               return_value='test') as mock_init:
            self.mock_docker.remove_container = mock.Mock(
                side_effect=errors.APIError('Error', '', ''))
            mock_container = mock.MagicMock()
            self.assertRaises(errors.APIError, self.driver.delete,
                              self.context, mock_container,
                              True)
            self.assertTrue(mock_cleanup_network_for_container.called)
            self.assertTrue(mock_cleanup_exposed_ports.called)
            self.mock_docker.remove_container.assert_called_once_with(
                mock_container.container_id, force=True)
            self.assertEqual(2, mock_init.call_count)

    def test_list(self):
        self.mock_docker.list_containers.return_value = []
        self.driver.list(self.context)
        self.mock_docker.list_containers.assert_called_once_with()

    def test_get_container_uuids(self):
        uuid = uuidutils.generate_uuid()
        uuid2 = uuidutils.generate_uuid()
        mock_container_list = [
            {'Names': ['/%s%s' % (consts.NAME_PREFIX, uuid)]},
            {'Names': ['/%s%s' % (consts.NAME_PREFIX, uuid2)]}]
        uuids = self.driver._get_container_uuids(mock_container_list)
        self.assertEqual(sorted([uuid, uuid2]), sorted(uuids))

    @mock.patch('zun.objects.container.Container.list')
    @mock.patch('zun.objects.container.Container.list_by_host')
    def test_get_local_containers(self, mock_list_by_host, mock_list):
        uuid = uuidutils.generate_uuid()
        uuid2 = uuidutils.generate_uuid()
        uuid3 = uuidutils.generate_uuid()
        mock_container = obj_utils.get_test_container(
            self.context, uuid=uuid, host='host')
        mock_container_2 = obj_utils.get_test_container(
            self.context, uuid=uuid2, host='host')
        mock_container_3 = obj_utils.get_test_container(
            self.context, uuid=uuid3, host='host2')

        def fake_container_list(context, filters):
            map = {}
            map[uuid] = mock_container
            map[uuid2] = mock_container_2
            map[uuid3] = mock_container_3
            return [map[u] for u in filters['uuid']]

        def fake_container_list_by_host(context, host):
            containers = [mock_container, mock_container_2, mock_container_3]
            return [c for c in containers if c.host == conf.CONF.host]

        mock_list.side_effect = fake_container_list
        mock_list_by_host.side_effect = fake_container_list_by_host

        # Containers in Docker matches DB records
        conf.CONF.set_override('host', 'host')
        docker_uuids = [uuid, uuid2]
        local_containers = self.driver._get_local_containers(
            self.context, docker_uuids)
        self.assertEqual(2, len(local_containers))
        self.assertIn(mock_container, local_containers)
        self.assertIn(mock_container_2, local_containers)
        self.assertNotIn(mock_container_3, local_containers)

        # Containers in Docker doesn't match DB records
        conf.CONF.set_override('host', 'host')
        docker_uuids = [uuid2, uuid3]
        local_containers = self.driver._get_local_containers(
            self.context, docker_uuids)
        self.assertEqual(3, len(local_containers))
        self.assertIn(mock_container, local_containers)
        self.assertIn(mock_container_2, local_containers)
        self.assertIn(mock_container_3, local_containers)

        # Containers are recorded in DB but missing in Docker
        conf.CONF.set_override('host', 'host')
        docker_uuids = []
        local_containers = self.driver._get_local_containers(
            self.context, docker_uuids)
        self.assertEqual(2, len(local_containers))
        self.assertIn(mock_container, local_containers)
        self.assertIn(mock_container_2, local_containers)
        self.assertNotIn(mock_container_3, local_containers)

        # Containers are present in Docker but not recorded in DB
        conf.CONF.set_override('host', 'host3')
        docker_uuids = [uuid2, uuid3]
        local_containers = self.driver._get_local_containers(
            self.context, docker_uuids)
        self.assertEqual(2, len(local_containers))
        self.assertNotIn(mock_container, local_containers)
        self.assertIn(mock_container_2, local_containers)
        self.assertIn(mock_container_3, local_containers)

    @mock.patch('zun.objects.container.Container.save')
    def test_update_containers_states(self, mock_save):
        mock_container = obj_utils.get_test_container(
            self.context, status='Running', host='host1')
        mock_container_2 = obj_utils.get_test_container(
            self.context, status='Stopped')
        conf.CONF.set_override('host', 'host2')
        with mock.patch.object(self.driver, 'list') as mock_list:
            mock_list.return_value = ([mock_container_2], [])
            self.assertEqual(mock_container.host, 'host1')
            self.assertEqual(mock_container.status, 'Running')
            self.driver.update_containers_states(
                self.context, [mock_container], mock.Mock())
            self.assertEqual(mock_container.host, 'host2')
            self.assertEqual(mock_container.status, 'Stopped')

    def test_heal_with_rebuilding_container(self):
        mock_compute_manager = mock.Mock()
        mock_container = obj_utils.get_test_container(
            self.context, status='Running',
            auto_heal=True, task_state=None)
        self.driver.heal_with_rebuilding_container(
            self.context, mock_container, mock_compute_manager)
        mock_compute_manager.container_rebuild.assert_called_once_with(
            self.context, mock_container)

    @mock.patch('zun.compute.api.API.container_rebuild')
    def test_heal_with_rebuilding_exception(self, mock_container_rebuild):
        container = Container(self.context, **utils.get_test_container())
        container.status = consts.RUNNING
        mock_container_rebuild.side_effect = Exception
        self.driver.heal_with_rebuilding_container(
            self.context, container, mock.Mock())

    def test_show_success(self):
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'running'})
        mock_container = mock.MagicMock()
        self.driver.show(self.context, mock_container)
        self.mock_docker.inspect_container.assert_called_once_with(
            mock_container.container_id)

    def test_show_container_created(self):
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        mock_container = mock.MagicMock(task_state=None)
        mock_container.status = consts.CREATING
        self.driver.show(self.context, mock_container)
        self.mock_docker.inspect_container.assert_called_once_with(
            mock_container.container_id)
        self.assertEqual(consts.CREATED, mock_container.status)

    def test_show_container_create_failed(self):
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': 'created',
                          'Config': {'Cmd': ['fake_command']}})
        mock_container = mock.MagicMock()
        mock_container.status = consts.ERROR
        self.driver.show(self.context, mock_container)
        self.mock_docker.inspect_container.assert_called_once_with(
            mock_container.container_id)
        self.assertEqual(consts.ERROR, mock_container.status)

    def test_show_fail_container_id_is_none(self):
        mock_container = mock.MagicMock()
        mock_container.container_id = None
        result_container = self.driver.show(self.context, mock_container)
        self.assertIsNone(result_container.container_id)

    def test_show_fail_container_status_error(self):
        with mock.patch.object(errors.APIError, '__str__',
                               return_value='404 Not Found') as mock_init:
            self.mock_docker.inspect_container = mock.Mock(
                side_effect=errors.APIError('Error', '', ''))
            mock_container = mock.MagicMock(auto_remove=False)
            result_container = self.driver.show(self.context, mock_container)
            self.mock_docker.inspect_container.assert_called_once_with(
                mock_container.container_id)
            self.assertEqual(result_container.status,
                             consts.ERROR)
            self.assertEqual(2, mock_init.call_count)

    def test_show_container_status_error_stop(self):
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': {
                'Status': 'exited',
                'Error': 'Container start error.',
                'FinishedAt': '0001-01-01T00:00:00Z',
            }}
        )
        mock_container = mock.MagicMock(task_state=None)
        mock_container.status = 'Created'
        mock_container.Error = 'Container start error.'
        self.driver.show(self.context, mock_container)
        self.mock_docker.inspect_container.assert_called_once_with(
            mock_container.container_id
        )
        self.assertEqual(consts.STOPPED, mock_container.status)

    def test_show_container_status_error_unknown(self):
        self.mock_docker.inspect_container = mock.Mock(
            return_value={'State': {
                'Status': 'Unknown',
                'Error': 'Container run error.',
                'FinishedAt': '0001-01-01T00:00:00Z',
            }}
        )
        mock_container = mock.MagicMock(task_state=None)
        mock_container.status = 'Running'
        mock_container.Error = 'Container run error.'
        self.driver.show(self.context, mock_container)
        self.mock_docker.inspect_container.assert_called_once_with(
            mock_container.container_id
        )
        self.assertEqual(consts.UNKNOWN, mock_container.status)

    def test_show_status_deleting(self):
        with mock.patch.object(errors.APIError, '__str__',
                               return_value='404 Not Found') as mock_init:
            self.mock_docker.inspect_container = mock.Mock(
                side_effect=errors.APIError('Error', '', ''))
            mock_container = mock.MagicMock(status=consts.DELETING)
            result_container = self.driver.show(self.context, mock_container)
            self.mock_docker.inspect_container.assert_called_once_with(
                mock_container.container_id)
            self.assertEqual(result_container.status,
                             consts.DELETING)
            self.assertEqual(1, mock_init.call_count)

    def test_show_fail_api_error(self):
        with mock.patch.object(errors.APIError, '__str__',
                               return_value='test') as mock_init:
            self.mock_docker.inspect_container = mock.Mock(
                side_effect=errors.APIError('Error', '', ''))
            mock_container = mock.MagicMock()
            self.assertRaises(errors.APIError, self.driver.show,
                              self.context, mock_container)
            self.mock_docker.inspect_container.assert_called_once_with(
                mock_container.container_id)
            self.assertEqual(1, mock_init.call_count)

    def test_reboot(self):
        self.mock_docker.restart = mock.Mock()
        mock_container = mock.MagicMock()
        result_container = self.driver.reboot(
            self.context, mock_container, '30')
        self.mock_docker.restart.assert_called_once_with(
            mock_container.container_id, timeout=30)
        self.assertEqual(result_container.status,
                         consts.RUNNING)

    def test_stop(self):
        self.mock_docker.stop = mock.Mock()
        mock_container = mock.MagicMock()
        result_container = self.driver.stop(self.context, mock_container, '30')
        self.mock_docker.stop.assert_called_once_with(
            mock_container.container_id,
            timeout=30)
        self.assertEqual(result_container.status,
                         consts.STOPPED)

    def test_start(self):
        self.mock_docker.start = mock.Mock()
        mock_container = mock.MagicMock()
        result_container = self.driver.start(self.context, mock_container)
        self.mock_docker.start.assert_called_once_with(
            mock_container.container_id)
        self.assertEqual(result_container.status,
                         consts.RUNNING)

    def test_pause(self):
        self.mock_docker.pause = mock.Mock()
        mock_container = mock.MagicMock()
        result_container = self.driver.pause(self.context, mock_container)
        self.mock_docker.pause.assert_called_once_with(
            mock_container.container_id)
        self.assertEqual(result_container.status,
                         consts.PAUSED)

    def test_unpause(self):
        self.mock_docker.unpause = mock.Mock()
        mock_container = mock.MagicMock()
        result_container = self.driver.unpause(self.context, mock_container)
        self.mock_docker.unpause.assert_called_once_with(
            mock_container.container_id)
        self.assertEqual(result_container.status,
                         consts.RUNNING)

    def test_show_logs(self):
        self.mock_docker.logs = mock.Mock()
        mock_container = mock.MagicMock()
        self.driver.show_logs(self.context, mock_container)
        self.mock_docker.logs.assert_called_once_with(
            mock_container.container_id, True, True, False, False,
            'all', None)

    def test_execute_create(self):
        self.mock_docker.exec_create = mock.Mock(return_value={'Id': 'test'})
        mock_container = mock.MagicMock()
        exec_id = self.driver.execute_create(
            self.context, mock_container, 'ls')
        self.assertEqual('test', exec_id)
        self.mock_docker.exec_create.assert_called_once_with(
            mock_container.container_id, 'ls', stdin=False, tty=False)

    def test_execute_run(self):
        self.mock_docker.exec_start = mock.Mock(return_value='test')
        self.mock_docker.exec_inspect = mock.Mock(
            return_value={'ExitCode': 0})
        self.driver.execute_run('test', 'ls')
        self.mock_docker.exec_start.assert_called_once_with('test', False,
                                                            False, False)
        self.mock_docker.exec_inspect.assert_called_once()

    def test_kill_successful_signal_is_none(self):
        self.mock_docker.kill = mock.Mock()
        mock_container = mock.MagicMock()
        self.driver.kill(self.context, mock_container, signal=None)
        self.mock_docker.kill.assert_called_once_with(
            mock_container.container_id)

    def test_kill_successful_signal_is_not_none(self):
        self.mock_docker.kill = mock.Mock()
        mock_container = mock.MagicMock()
        self.driver.kill(self.context, mock_container, signal='test')
        self.mock_docker.kill.assert_called_once_with(
            mock_container.container_id,
            'test')

    def test_resize(self):
        self.mock_docker.resize = mock.Mock()
        mock_container = mock.MagicMock()
        self.driver.resize(self.context, mock_container, "100", "100")
        self.mock_docker.resize.assert_called_once_with(
            mock_container.container_id, 100, 100)

    def test_commit(self):
        self.mock_docker.commit = mock.Mock()
        mock_container = mock.MagicMock()
        self.driver.commit(self.context, mock_container, "repo", "tag")
        self.mock_docker.commit.assert_called_once_with(
            mock_container.container_id, "repo", "tag")

    def test_execute_resize(self):
        self.mock_docker.exec_resize = mock.Mock()
        fake_exec_id = 'fake_id'
        self.driver.execute_resize(fake_exec_id, "100", "100")
        self.mock_docker.exec_resize.assert_called_once_with(
            fake_exec_id, height=100, width=100)

    def test_get_host_info(self):
        self.mock_docker.info = mock.Mock()
        self.mock_docker.info.return_value = {'Containers': 10,
                                              'ContainersPaused': 0,
                                              'ContainersRunning': 8,
                                              'ContainersStopped': 2,
                                              'NCPU': 48,
                                              'Architecture': 'x86_64',
                                              'OSType': 'linux',
                                              'OperatingSystem': 'CentOS',
                                              'KernelVersion': '3.10.0-123',
                                              'Labels': ['dev.type=product'],
                                              'Runtimes': {'runc': {'path':
                                                           'docker-runc'}},
                                              'DockerRootDir': 'fake-dir'}
        host_info = self.driver.get_host_info()
        self.assertEqual(10, host_info['total_containers'])
        self.assertEqual(8, host_info['running_containers'])
        self.assertEqual(0, host_info['paused_containers'])
        self.assertEqual(2, host_info['stopped_containers'])
        self.assertEqual(48, host_info['cpus'])
        self.assertEqual('x86_64', host_info['architecture'])
        self.assertEqual('linux', host_info['os_type'])
        self.assertEqual('CentOS', host_info['os'])
        self.assertEqual('3.10.0-123', host_info['kernel_version'])
        self.assertEqual({"dev.type": "product"}, host_info['labels'])
        self.assertEqual('fake-dir', host_info['docker_root_dir'])

    def test_stats(self):
        self.mock_docker.stats = mock.Mock()
        mock_container = mock.MagicMock()
        self.mock_docker.stats.return_value = {
            'cpu_stats': {'cpu_usage': {'usage_in_usermode': 1000000000,
                                        'total_usage': 1000000000},
                          'system_cpu_usage': 1000000000000},
            'blkio_stats': {'io_service_bytes_recursive':
                            [{'major': 253, 'value': 10000000,
                              'minor': 4, 'op': 'Read'},
                             {'major': 253, 'value': 0,
                              'minor': 4, 'op': 'Write'},
                             {'major': 253, 'value': 0,
                              'minor': 4, 'op': 'Sync'},
                             {'major': 253, 'value': 10000000,
                              'minor': 4, 'op': 'Async'},
                             {'major': 253, 'value': 10000000,
                              'minor': 4, 'op': 'Total'}]},
            'memory_stats': {'usage': 104857600,
                             'limit': 1048576000},
            'networks': {'eth0':
                         {'tx_dropped': 0, 'rx_packets': 2, 'rx_bytes': 200,
                             'tx_errors': 0, 'rx_errors': 0, 'tx_bytes': 200,
                             'rx_dropped': 0, 'tx_packets': 2}}}
        stats_info = self.driver.stats(self.context, mock_container)
        self.assertEqual(0.1, stats_info['CPU %'])
        self.assertEqual(100, stats_info['MEM USAGE(MiB)'])
        self.assertEqual(1000, stats_info['MEM LIMIT(MiB)'])
        self.assertEqual(10, stats_info['MEM %'])
        self.assertEqual('10000000/0', stats_info['BLOCK I/O(B)'])
        self.assertEqual('200/200', stats_info['NET I/O(B)'])

    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.disconnect_container_from_network')
    def test_network_detach(self, mock_detach):
        mock_container = mock.MagicMock()
        self.driver.network_detach(self.context, mock_container, 'network')
        mock_detach.assert_called_once_with(mock_container, 'network')

    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.disconnect_container_from_network')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.get_or_create_network')
    def test_network_attach(self, mock_get_or_create, mock_disconnect,
                            mock_connect):
        mock_container = mock.Mock()
        mock_container.security_groups = None
        mock_container.addresses = {}
        requested_network = {'network': 'network',
                             'port': '',
                             'fixed_ip': '',
                             'preserve_on_delete': False}
        self.driver.network_attach(self.context, mock_container,
                                   requested_network)
        mock_connect.assert_called_once_with(mock_container,
                                             requested_network,
                                             security_groups=None)

    def test_network_attach_error(self):
        mock_container = mock.Mock()
        mock_container.security_groups = None
        mock_container.addresses = {'already-attached-net': []}
        requested_network = {'network': 'already-attached-net',
                             'port': '',
                             'fixed_ip': '',
                             'preserve_on_delete': False}
        self.assertRaises(exception.ZunException,
                          self.driver.network_attach,
                          self.context, mock_container, requested_network)

    @mock.patch('zun.common.utils.get_security_group_ids')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.connect_container_to_network')
    @mock.patch('zun.network.kuryr_network.KuryrNetwork'
                '.get_or_create_network')
    def test_network_attach_with_security_group(self, mock_get_or_create,
                                                mock_connect,
                                                mock_get_sec_group_id):
        test_sec_group_id = '84e3a4c1-c8cd-46b1-a0d9-c8c35f6a32a4'
        mock_container = mock.Mock()
        mock_container.security_groups = ['test_sec_group']
        mock_container.addresses = {}
        mock_get_sec_group_id.return_value = test_sec_group_id
        requested_network = {'network': 'network',
                             'port': '',
                             'fixed_ip': '',
                             'preserve_on_delete': False}
        self.driver.network_attach(self.context, mock_container,
                                   requested_network)
        mock_connect.assert_called_once_with(mock_container,
                                             requested_network,
                                             security_groups=test_sec_group_id)

    @mock.patch('zun.common.utils.execute')
    @mock.patch('zun.container.os_capability.linux.os_capability_linux'
                '.LinuxHost.get_mem_numa_info')
    @mock.patch('zun.container.os_capability.linux.os_capability_linux'
                '.LinuxHost.get_cpu_numa_info')
    @mock.patch('zun.container.docker.driver.DockerDriver'
                '.get_total_disk_for_container')
    @mock.patch('zun.container.docker.driver.DockerDriver'
                '.get_host_mem')
    @mock.patch(
        'zun.container.docker.driver.DockerDriver.get_host_info')
    def test_get_available_resources(self, mock_info, mock_mem,
                                     mock_disk, mock_numa_cpu, mock_numa_mem,
                                     mock_output):
        self.driver = DockerDriver()
        numa_cpu_info = defaultdict(list)
        numa_cpu_info['0'] = [0, 8]
        mock_numa_cpu.return_value = numa_cpu_info
        mock_numa_mem.return_value = [1024 * 32]
        mock_output.return_value = LSCPU_ON
        conf.CONF.set_override('floating_cpu_set', "0")
        mock_mem.return_value = (100 * units.Ki, 50 * units.Ki, 50 * units.Ki,
                                 50 * units.Ki)
        mock_info.return_value = {'total_containers': 10,
                                  'running_containers': 8,
                                  'paused_containers': 0,
                                  'stopped_containers': 2,
                                  'cpus': 48,
                                  'architecture': 'x86_64',
                                  'os_type': 'linux',
                                  'os': 'CentOS',
                                  'kernel_version': '3.10.0-123',
                                  'labels': {'dev.type': 'product'},
                                  'runtimes': ['runc'],
                                  'enable_cpu_pinning': False,
                                  'docker_root_dir': '/var/lib/docker'}
        mock_disk.return_value = (100, 20)
        data = self.driver.get_available_resources()
        self.assertEqual(_numa_topo_spec, data['numa_topology'].to_list())
        self.assertEqual(100, data['mem_total'])
        self.assertEqual(50, data['mem_free'], 50)
        self.assertEqual(50, data['mem_available'])
        self.assertEqual(10, data['total_containers'])
        self.assertEqual(8, data['running_containers'])
        self.assertEqual(0, data['paused_containers'])
        self.assertEqual(2, data['stopped_containers'])
        self.assertEqual(48, data['cpus'])
        self.assertEqual('x86_64', data['architecture'])
        self.assertEqual('linux', data['os_type'])
        self.assertEqual('CentOS', data['os'])
        self.assertEqual('3.10.0-123', data['kernel_version'])
        self.assertEqual({'dev.type': 'product'}, data['labels'])
        self.assertEqual(80, data['disk_total'])
        self.assertEqual(['runc'], data['runtimes'])
